# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import List, Optional, Tuple
from lean.components.api.api_client import APIClient
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.util.library_manager import LibraryManager
from lean.components.util.logger import Logger
from lean.components.util.platform_manager import PlatformManager
from lean.components.util.project_manager import ProjectManager
from lean.models.api import QCProject, QCLanguage, QCProjectLibrary
from lean.components.util.organization_manager import OrganizationManager
from lean.models.errors import RequestFailedError
from lean.models.utils import LeanLibraryReference
from lean.models.encryption import ActionType
from lean.components.config.storage import safe_save

class PullManager:
    """The PullManager class is responsible for synchronizing cloud projects to the local drive."""

    def __init__(self,
                 logger: Logger,
                 api_client: APIClient,
                 project_manager: ProjectManager,
                 project_config_manager: ProjectConfigManager,
                 library_manager: LibraryManager,
                 platform_manager: PlatformManager,
                 organization_manager: OrganizationManager) -> None:
        """Creates a new PullManager instance.

        :param logger: the logger to use when printing messages
        :param api_client: the APIClient instance to use when communicating with the cloud
        :param project_manager: the ProjectManager instance to use when creating new projects
        :param project_config_manager: the ProjectConfigManager instance to use
        :param library_manager: the LibraryManager instance to use when updating library references
        :param platform_manager: the PlatformManager used when checking which operating system is in use
        """
        self._logger = logger
        self._api_client = api_client
        self._project_manager = project_manager
        self._project_config_manager = project_config_manager
        self._library_manager = library_manager
        self._platform_manager = platform_manager
        self._organization_manager = organization_manager
        self._last_file = None

    def _get_libraries(self, project: QCProject,
                       seen_projects: List[int] = None) -> Tuple[List[QCProject], List[QCProjectLibrary]]:
        """Gets the libraries referenced by the given project.

        :return: two lists including the libraries referenced by the project.
            The first one containing the library projects that could be fetched and
            the second list containing the libraries that could not be fetched because the user has no access to them.
        """
        if seen_projects is None:
            seen_projects = [project.projectId]

        libraries = []
        inaccessible_libraries = []
        for library in project.libraries:
            if library.projectId in seen_projects:
                continue

            if not library.access:
                inaccessible_libraries.append(library)
                continue

            seen_projects.append(library.projectId)

            try:
                library_project = self._api_client.projects.get(library.projectId, project.organizationId)
            except RequestFailedError:
                # the library could not be fetched, probably because it was deleted
                inaccessible_libraries.append(library)
                continue

            libraries.append(library_project)

            libraries_for_library, inaccessible_libs_for_library = self._get_libraries(library_project, seen_projects)
            libraries.extend(libraries_for_library)
            inaccessible_libraries.extend(inaccessible_libs_for_library)

        return libraries, inaccessible_libraries

    def pull_projects(self, projects_to_pull: List[QCProject], all_cloud_projects: Optional[List[QCProject]] = None, encryption_action: Optional[ActionType]=None, encryption_key: Optional[Path]=None) -> None:
        """Pulls the given projects from the cloud to the local drive.

        This will also pull libraries referenced by the project.

        :param projects_to_pull: the cloud projects that need to be pulled
        :param all_cloud_projects: all the projects available in the cloud
        """
        if all_cloud_projects is not None:
            projects, inaccessible_libraries = self._project_manager.get_cloud_projects_libraries(all_cloud_projects,
                                                                                               projects_to_pull)
            projects_to_pull.extend(projects)
        else:
            inaccessible_libraries = []
            for project in projects_to_pull:
                projects, libs_not_found = self._get_libraries(project, [p.projectId for p in projects_to_pull])
                projects_to_pull.extend(projects)
                inaccessible_libraries.extend(libs_not_found)

        for library in inaccessible_libraries:
            # let's build the right message, either the user has no access to the library or it may have been deleted
            projects_referencing_library = [p for p in projects_to_pull
                                            if any(lib.projectId == library.projectId for lib in p.libraries)]
            if len(projects_referencing_library) == 1:
                projects_str = f"project '{projects_referencing_library[0].name}'"
            else:
                joined_projects = ", ".join([f"''{p.name}''" for p in projects_referencing_library])
                projects_str = f"projects {joined_projects}"

            if library.access:
                reason = "The library might haven been deleted but the project still references it."
            else:
                reason = f"You don't have access to the library. " \
                         f"Please ask '{library.ownerName}' to add you as a collaborator to the project "\
                         f"in order to pull it."
            self._logger.warn(f"Cannot pull library '{library.libraryName}', which is referenced by {projects_str}. "
                              + reason)

        projects_to_pull = sorted(projects_to_pull, key=lambda p: p.name)
        projects_with_paths = []

        for index, project in enumerate(projects_to_pull, start=1):
            try:
                self._logger.info(f"[{index}/{len(projects_to_pull)}] Pulling '{project.name}'")
                projects_with_paths.append((project, self._pull_project(project, encryption_action, encryption_key)))
            except Exception as ex:
                from traceback import format_exc
                self._logger.debug(format_exc().strip())
                if self._last_file is not None:
                    self._logger.warn(
                        f"Cannot pull '{project.name}' (id {project.projectId}, failed on {self._last_file}): {ex}")
                else:
                    self._logger.warn(f"Cannot pull '{project.name}' (id {project.projectId}): {ex}")

        self._update_local_library_references(projects_with_paths)

    def _pull_project(self, project: QCProject, encryption_action: Optional[ActionType], encryption_key: Optional[Path]) -> Path:
        """Pulls a single project from the cloud to the local drive.

        Raises an error with a descriptive message if the project cannot be pulled.

        :param project: the cloud project to pull
        :return the actual local path of the project
        """
        local_project_path = self._project_manager.get_local_project_path(project.name, project.projectId,
                                                                          allow_corrupted=True)
        local_project_name = local_project_path.relative_to(Path.cwd()).as_posix()
        # Check if cloud project has invalid name, if so update it and inform user.
        if local_project_name != project.name:
            # update project name in cloud
            self._api_client.projects.update(project.projectId, **{"name": local_project_name})
            self._logger.info(f"Renamed project in cloud from '{project.name}' to '{local_project_name}'")
            project.name = local_project_name

        # rename project on disk if we find a directory with the old name (invalid/renamed name)
        # only check for old directory if expected directory does not exist
        if not local_project_path.exists():
            project_path_on_disk = self._project_manager.try_get_project_path_by_cloud_id(project.projectId)
            if project_path_on_disk:
                project_name_on_disk = project_path_on_disk.relative_to(Path.cwd()).as_posix()
                if project_name_on_disk != project.name:
                    self._project_manager.rename_project_and_contents(project_path_on_disk, Path.cwd() / project.name)

        project_config = self._project_config_manager.get_project_config(local_project_path)
        local_encryption_state = project_config.get("encrypted", False)
        local_encryption_key = project_config.get("encryption-key-path", None)
        if local_encryption_key is not None:
            local_encryption_key = Path(local_encryption_key)
        # Handle mismatch cases
        from lean.components.util.encryption_helper import validate_key_and_encryption_state_for_cloud_project
        validate_key_and_encryption_state_for_cloud_project(project, local_encryption_state, encryption_key, local_encryption_key, self._logger)

        # Pull the cloud files to the local drive
        self._pull_files(project, local_project_path, encryption_action, encryption_key)

        # Update the local project config with the latest details
        project_config = self._project_config_manager.get_project_config(local_project_path)
        project_config.set("cloud-id", project.projectId)
        project_config.set("algorithm-language", project.language.name)
        project_config.set("parameters", {parameter.key: parameter.value for parameter in project.parameters})
        project_config.set("description", project.description)
        project_config.set("organization-id", project.organizationId)
        project_config.set("python-venv", project.leanEnvironment)
        if encryption_key:
            project_config.set("encrypted", encryption_action == ActionType.ENCRYPT)
        else:
            project_config.set('encrypted', project.encrypted)

        if not project.leanPinnedToMaster:
            project_config.set("lean-engine", project.leanVersionId)
        else:
            project_config.delete("lean-engine")

        return local_project_path

    def _pull_files(self, project: QCProject, local_project_path: Path, encryption_action: Optional[ActionType], encryption_key: Optional[Path]) -> None:
        """Pull the files of a single project.

        :param project: the cloud project of which the files need to be pulled
        :param local_project_path: the path to the local project directory
        """
        if not local_project_path.exists():
            self._project_manager.create_new_project(local_project_path, project.language)

        cloud_files = self._api_client.files.get_all(project.projectId)
        if encryption_key:
            from lean.components.util.encryption_helper import get_appropriate_files_from_cloud_project
            organization_id = self._organization_manager.try_get_working_organization_id()
            cloud_files = get_appropriate_files_from_cloud_project(project, cloud_files, encryption_key, organization_id, encryption_action)

        for cloud_file in cloud_files:
            self._last_file = cloud_file.name

            if cloud_file.isLibrary:
                continue

            local_file_path = local_project_path / cloud_file.name

            # Skip if the local file already exists with the correct content
            if local_file_path.exists():
                if local_file_path.read_text(encoding="utf-8").strip() == cloud_file.content.strip():
                    self._project_manager.update_last_modified_time(local_file_path, cloud_file.modified)
                    continue

            local_file_path.parent.mkdir(parents=True, exist_ok=True)

            # Make sure we always work with unix line endings in memory,
            # so they can be properly translated to the local OS line endings when writing to disk.
            content = cloud_file.content.replace("\r\n", "\n")
            if content != "" and not content.endswith("\n"):
                content += "\n"
            safe_save(content, local_file_path)

            self._project_manager.update_last_modified_time(local_file_path, cloud_file.modified)
            self._logger.info(f"Successfully pulled '{project.name}/{cloud_file.name}'")

        self._last_file = None
        self._project_manager.update_last_modified_time(local_project_path, project.modified)

    def _add_local_library_references_to_project(self, project_dir: Path, cloud_libraries_paths: List[Path]) -> None:
        if len(cloud_libraries_paths) > 0:
            self._logger.info(f"Adding/updating local library references to project {project_dir.name}")

        for i, library_dir in enumerate(cloud_libraries_paths, start=1):
            self._logger.info(f"[{i}/{len(cloud_libraries_paths)}] Adding/updating local library "
                              f"{library_dir.name} reference to project {project_dir.name}")
            # Add library references without restoring. It will be done after all lib references have been updated
            self._library_manager.add_lean_library_to_project(project_dir, library_dir, True)

    def _remove_local_library_references_from_project(self,
                                                      project_dir: Path,
                                                      cloud_libraries_paths: List[Path]) -> None:

        project_config = self._project_config_manager.get_project_config(project_dir)
        local_libraries = project_config.get("libraries", [])
        cloud_library_relative_paths = [library_dir.relative_to(Path.cwd()) for library_dir in cloud_libraries_paths]
        libraries_to_remove = [LeanLibraryReference(**library_reference)
                               for library_reference in local_libraries
                               if Path(library_reference["path"]) not in cloud_library_relative_paths]

        if len(libraries_to_remove) > 0:
            self._logger.info(f"Removing local library references from project {project_dir.name}")

        for i, library_reference in enumerate(libraries_to_remove, start=1):
            self._logger.info(f"[{i}/{len(libraries_to_remove)}] Removing local library "
                              f"{library_reference.name} reference from project {project_dir.name}")
            # Remove library references without restoring. It will be done after all lib references have been updated
            self._library_manager.remove_lean_library_from_project(project_dir,
                                                                   Path.cwd() / library_reference.path,
                                                                   True)

    def _update_local_library_references(self, projects: List[Tuple[QCProject, Path]]) -> None:
        for project, path in projects:
            libraries = [library.projectId for library in project.libraries]
            cloud_libraries_paths = [library_path for library, library_path in projects
                                     if library.projectId in libraries]

            # Add cloud library references to local config
            self._add_local_library_references_to_project(path, cloud_libraries_paths)

            # Remove library references locally if they were removed in the cloud
            self._remove_local_library_references_from_project(path, cloud_libraries_paths)

            # Restore the project to automatically enable local auto-complete
            try:
                self._restore_project(project, path)
            except RuntimeWarning as e:
                self._logger.info(e)

    def _restore_project(self, project: QCProject, project_dir: Path) -> None:
        if project.language != QCLanguage.CSharp:
            return

        project_csproj_file = self._project_manager.get_csproj_file_path(project_dir)
        self._project_manager.try_restore_csharp_project(project_csproj_file)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple
from click import option, argument, Choice
from lean.click import LeanCommand, PathParameter
from lean.components.util.name_rename import rename_internal_config_to_user_friendly_format
from lean.constants import DEFAULT_ENGINE_IMAGE
from lean.container import container
from lean.models.cli import (cli_brokerages, cli_data_queue_handlers, cli_data_downloaders,
                             cli_addon_modules, cli_history_provider)
from lean.models.errors import MoreInfoError
from lean.models.click_options import options_from_json, get_configs_for_options
from lean.models.json_module import LiveInitialStateInput, JsonModule
from lean.commands.live.live import live
from lean.components.util.live_utils import get_last_portfolio_cash_holdings, configure_initial_cash_balance, configure_initial_holdings,\
                                            _configure_initial_cash_interactively, _configure_initial_holdings_interactively
from lean.components.util.json_modules_handler import build_and_configure_modules, \
    non_interactive_config_build_for_name, interactive_config_build, config_build_for_name

_environment_skeleton = {
    "live-mode": True,
    "setup-handler": "QuantConnect.Lean.Engine.Setup.BrokerageSetupHandler",
    "result-handler": "QuantConnect.Lean.Engine.Results.LiveTradingResultHandler",
    "data-feed-handler": "QuantConnect.Lean.Engine.DataFeeds.LiveTradingDataFeed",
    "real-time-handler": "QuantConnect.Lean.Engine.RealTime.LiveTradingRealTimeHandler",
    "history-provider": ["SubscriptionDataReaderHistoryProvider"],
    "transaction-handler": "QuantConnect.Lean.Engine.TransactionHandlers.BrokerageTransactionHandler"
}


def _start_iqconnect_if_necessary(lean_config: Dict[str, Any], environment_name: str) -> None:
    """Starts IQConnect if the given environment uses IQFeed as data queue handler.

    :param lean_config: the LEAN configuration that should be used
    :param environment_name: the name of the environment
    """
    from subprocess import Popen

    environment = lean_config["environments"][environment_name]
    if environment["data-queue-handler"] != "QuantConnect.ToolBox.IQFeed.IQFeedDataQueueHandler":
        return

    args = [lean_config["iqfeed-iqconnect"],
            "-product", lean_config["iqfeed-productName"],
            "-version", lean_config["iqfeed-version"]]

    username = lean_config.get("iqfeed-username", "")
    if username != "":
        args.extend(["-login", username])

    password = lean_config.get("iqfeed-password", "")
    if password != "":
        args.extend(["-password", password])

    Popen(args)

    container.logger.info("Waiting 10 seconds for IQFeed to start")
    from time import sleep
    sleep(10)


def _try_get_data_downloader_name(data_downloader_name: str, data_provider_live_names: [str]) -> str:
    """ Get name for data downloader provider based on data provider live (if exist)

    :param data_downloader_name: the current (default) data provider historical
    :param data_provider_live_names: the current data provider live names
    """
    return next((live_data_historical.get_name() for live_data_historical in cli_data_downloaders
                 if live_data_historical.get_name() in data_provider_live_names), data_downloader_name)


def _get_history_provider_name(data_provider_live_names: [str]) -> [str]:
    """ Get name for history providers based on the live data providers

    :param data_provider_live_names: the current data provider live names
    """
    history_providers = []
    for potential_history_provider in cli_history_provider:
        if potential_history_provider.get_name() in data_provider_live_names:
            history_providers.append(potential_history_provider.get_name())
    return history_providers


@live.command(cls=LeanCommand, requires_lean_config=True, requires_docker=True, default_command=True, name="deploy")
@argument("project", type=PathParameter(exists=True, file_okay=True, dir_okay=True))
@option("--environment",
              type=str,
              help="The environment to use")
@option("--output",
              type=PathParameter(exists=False, file_okay=False, dir_okay=True),
              help="Directory to store results in (defaults to PROJECT/live/TIMESTAMP)")
@option("--detach", "-d",
              is_flag=True,
              default=False,
              help="Run the live deployment in a detached Docker container and return immediately")
@option("--brokerage",
              type=Choice([b.get_name() for b in cli_brokerages], case_sensitive=False),
              help="The brokerage to use")
@option("--data-provider-live",
              type=Choice([d.get_name() for d in cli_data_queue_handlers], case_sensitive=False),
              multiple=True,
              help="The live data provider to use")
@option("--data-provider-historical",
              type=Choice([dp.get_name() for dp in cli_data_downloaders if dp.get_id() != "TerminalLinkBrokerage"], case_sensitive=False),
              help="Update the Lean configuration file to retrieve data from the given historical provider")
@options_from_json(get_configs_for_options("live-cli"))
@option("--release",
              is_flag=True,
              default=False,
              help="Compile C# projects in release configuration instead of debug")
@option("--image",
              type=str,
              help=f"The LEAN engine image to use (defaults to {DEFAULT_ENGINE_IMAGE})")
@option("--python-venv",
              type=str,
              help=f"The path of the python virtual environment to be used")
@option("--live-cash-balance",
              type=str,
              default="",
              help=f"A comma-separated list of currency:amount pairs of initial cash balance")
@option("--live-holdings",
              type=str,
              default="",
              help=f"A comma-separated list of symbol:symbolId:quantity:averagePrice of initial portfolio holdings")
@option("--update",
              is_flag=True,
              default=False,
              help="Pull the LEAN engine image before starting live trading")
@option("--show-secrets", is_flag=True, show_default=True, default=False, help="Show secrets as they are input")
@option("--addon-module",
              type=str,
              multiple=True,
              hidden=True)
@option("--extra-config",
              type=(str, str),
              multiple=True,
              hidden=True)
@option("--extra-docker-config",
              type=str,
              default="{}",
              help="Extra docker configuration as a JSON string. "
                   "For more information https://docker-py.readthedocs.io/en/stable/containers.html")
@option("--no-update",
              is_flag=True,
              default=False,
              help="Use the local LEAN engine image instead of pulling the latest version")
def deploy(project: Path,
           environment: Optional[str],
           output: Optional[Path],
           detach: bool,
           brokerage: Optional[str],
           data_provider_live: Optional[str],
           data_provider_historical: Optional[str],
           release: bool,
           image: Optional[str],
           python_venv: Optional[str],
           live_cash_balance: Optional[str],
           live_holdings: Optional[str],
           update: bool,
           show_secrets: bool,
           addon_module: Optional[List[str]],
           extra_config: Optional[Tuple[str, str]],
           extra_docker_config: Optional[str],
           no_update: bool,
           **kwargs) -> None:
    """Start live trading a project locally using Docker.

    \b
    If PROJECT is a directory, the algorithm in the main.py or Main.cs file inside it will be executed.
    If PROJECT is a file, the algorithm in the specified file will be executed.

    By default an interactive wizard is shown letting you configure the brokerage and live data provider to use.
    If --environment, --brokerage or --data-provider-live are given the command runs in non-interactive mode.
    In this mode the CLI does not prompt for input.

    If --environment is given it must be the name of a live environment in the Lean configuration.

    If --brokerage and --data-provider-live are given, the options specific to the given brokerage/live data provider must also be given.
    The Lean config is used as fallback when a brokerage/live data provider-specific option hasn't been passed in.
    If a required option is not given and cannot be found in the Lean config the command aborts.

    By default the official LEAN engine image is used.
    You can override this using the --image option.
    Alternatively you can set the default engine image for all commands using `lean config set engine-image <image>`.
    """
    from copy import copy
    from datetime import datetime
    from json import loads

    logger = container.logger

    project_manager = container.project_manager
    algorithm_file = project_manager.find_algorithm_file(Path(project))

    if output is None:
        output = algorithm_file.parent / "live" / datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    lean_config_manager = container.lean_config_manager

    brokerage_instance: JsonModule
    data_provider_live_instances: [JsonModule] = []
    history_providers: [str] = []
    history_providers_instances: [JsonModule] = []
    data_downloader_instances: JsonModule
    if environment is not None and (brokerage is not None or len(data_provider_live) > 0):
        raise RuntimeError("--environment and --brokerage + --data-provider-live are mutually exclusive")

    environment_name = "lean-cli"
    if environment is not None:
        environment_name = environment
        lean_config = lean_config_manager.get_complete_lean_config(environment_name, algorithm_file, None)

        if environment_name in lean_config["environments"]:
            lean_environment = lean_config["environments"][environment_name]
            for key in ["live-mode-brokerage", "data-queue-handler", "history-provider"]:
                if key not in lean_environment:
                    raise MoreInfoError(f"The '{environment_name}' environment does not specify a {rename_internal_config_to_user_friendly_format(key)}",
                                        "https://www.lean.io/docs/v2/lean-cli/live-trading/algorithm-control")

            brokerage = lean_environment["live-mode-brokerage"]
            data_provider_live = lean_environment["data-queue-handler"]
            if type(data_provider_live) is not list:
                data_provider_live = [data_provider_live]
            history_providers = lean_environment["history-provider"]
            if type(history_providers) is not list:
                history_providers = [history_providers]
            logger.debug(f'Deploy(): loading env \'{environment_name}\'. Brokerage: \'{brokerage}\'. IDQHs: '
                         f'{data_provider_live}. HistoryProviders: {history_providers}')
        else:
            logger.info(f'Environment \'{environment_name}\' not found, creating from scratch')
            lean_config["environments"] = {environment_name: copy(_environment_skeleton)}
    else:
        lean_config = lean_config_manager.get_complete_lean_config(environment_name, algorithm_file, None)
        lean_config["environments"] = {environment_name: copy(_environment_skeleton)}

    if brokerage:
        # user provided brokerage, check all arguments were provided
        brokerage_instance = non_interactive_config_build_for_name(lean_config, brokerage, cli_brokerages, kwargs,
                                                                   logger, environment_name)
    else:
        # let the user choose the brokerage
        brokerage_instance = interactive_config_build(lean_config, cli_brokerages, logger, kwargs, show_secrets,
                                                      "Select a brokerage", multiple=False,
                                                      environment_name=environment_name)

    if data_provider_live and len(data_provider_live) > 0:
        for data_feed_name in data_provider_live:
            data_feed = non_interactive_config_build_for_name(lean_config, data_feed_name, cli_data_queue_handlers,
                                                              kwargs, logger, environment_name)
            data_provider_live_instances.append(data_feed)
    else:
        data_provider_live_instances = interactive_config_build(lean_config, cli_data_queue_handlers, logger, kwargs,
                                                                show_secrets, "Select a live data feed", multiple=True,
                                                                environment_name=environment_name)

    # based on the live data providers we set up the history providers
    data_provider_live = [provider.get_name() for provider in data_provider_live_instances]
    if data_provider_historical is None:
        data_provider_historical = _try_get_data_downloader_name("Local", data_provider_live)
    data_downloader_instances = non_interactive_config_build_for_name(lean_config, data_provider_historical,
                                                                      cli_data_downloaders, kwargs, logger,
                                                                      environment_name)
    if history_providers is None or len(history_providers) == 0:
        history_providers = _get_history_provider_name(data_provider_live)
    for history_provider in history_providers:
        if history_provider in ["BrokerageHistoryProvider", "SubscriptionDataReaderHistoryProvider"]:
            continue
        history_providers_instances.append(config_build_for_name(lean_config, history_provider, cli_history_provider,
                                                                 kwargs, logger, interactive=True,
                                                                 environment_name=environment_name))

    organization_id = container.organization_manager.try_get_working_organization_id()
    paths_to_mount = {}
    for module in (data_provider_live_instances + [data_downloader_instances, brokerage_instance]
                   + history_providers_instances):
        module.ensure_module_installed(organization_id)
        paths_to_mount.update(module.get_paths_to_mount())

    if not lean_config["environments"][environment_name]["live-mode"]:
        raise MoreInfoError(f"The '{environment_name}' is not a live trading environment (live-mode is set to false)",
                            "https://www.lean.io/docs/v2/lean-cli/live-trading/brokerages/quantconnect-paper-trading")

    project_config_manager = container.project_config_manager
    cli_config_manager = container.cli_config_manager

    project_config = project_config_manager.get_project_config(algorithm_file.parent)
    engine_image = cli_config_manager.get_engine_image(image or project_config.get("engine-image", None))

    container.update_manager.pull_docker_image_if_necessary(engine_image, update, no_update)

    _start_iqconnect_if_necessary(lean_config, environment_name)

    if not output.exists():
        output.mkdir(parents=True)

    if python_venv is not None and python_venv != "":
        lean_config["python-venv"] = f'{"/" if python_venv[0] != "/" else ""}{python_venv}'

    cash_balance_option, holdings_option, last_cash, last_holdings = get_last_portfolio_cash_holdings(container.api_client, brokerage_instance,
                                                                                                      project_config.get("cloud-id", None), project)

    if environment is None and brokerage is None and len(data_provider_live) == 0:   # condition for using interactive panel
        if cash_balance_option != LiveInitialStateInput.NotSupported:
            live_cash_balance = _configure_initial_cash_interactively(logger, cash_balance_option, last_cash)

        if holdings_option != LiveInitialStateInput.NotSupported:
            live_holdings = _configure_initial_holdings_interactively(logger, holdings_option, last_holdings)
    else:
        if cash_balance_option != LiveInitialStateInput.NotSupported:
            live_cash_balance = configure_initial_cash_balance(logger, cash_balance_option, live_cash_balance, last_cash)
        elif live_cash_balance is not None and live_cash_balance != "":
            raise RuntimeError(f"Custom cash balance setting is not available for {brokerage_instance}")

        if holdings_option != LiveInitialStateInput.NotSupported:
            live_holdings = configure_initial_holdings(logger, holdings_option, live_holdings, last_holdings)
        elif live_holdings is not None and live_holdings != "":
            raise RuntimeError(f"Custom portfolio holdings setting is not available for {brokerage_instance}")

    if live_cash_balance:
        lean_config["live-cash-balance"] = live_cash_balance
    if live_holdings:
        lean_config["live-holdings"] = [{
            "Symbol": {
                "Value": holding["symbol"],
                "ID": holding["symbolId"]
            },
            "Quantity": holding["quantity"],
            "AveragePrice": holding["averagePrice"]
        } for holding in live_holdings]

    if str(engine_image) != DEFAULT_ENGINE_IMAGE:
        logger.warn(f'A custom engine image: "{engine_image}" is being used!')

    # Set extra config
    given_algorithm_id = None
    for key, value in extra_config:
        if key == "algorithm-id":
            given_algorithm_id = int(value)
        else:
            lean_config[key] = value

    output_config_manager = container.output_config_manager
    lean_config["algorithm-id"] = f"L-{output_config_manager.get_live_deployment_id(output, given_algorithm_id)}"

    # Configure addon modules
    build_and_configure_modules(addon_module, cli_addon_modules, organization_id, lean_config,
                                kwargs, logger, environment_name)

    if container.platform_manager.is_host_arm():
        if "InteractiveBrokersBrokerage" in lean_config["environments"][environment_name]["live-mode-brokerage"] \
                or any("InteractiveBrokersBrokerage" in dataQueue for dataQueue in lean_config["environments"][environment_name]["data-queue-handler"]):
            raise RuntimeError(f"InteractiveBrokers is currently not supported for ARM hosts")

    lean_runner = container.lean_runner
    lean_runner.run_lean(lean_config,
                         environment_name,
                         algorithm_file,
                         output,
                         engine_image,
                         None,
                         release,
                         detach,
                         loads(extra_docker_config),
                         paths_to_mount)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Any, Dict, List, Type

from click import get_current_context
from click.core import ParameterSource

from lean.components.util.logger import Logger
from lean.constants import MODULE_TYPE, MODULE_PLATFORM, MODULE_CLI_PLATFORM
from lean.container import container
from lean.models.configuration import BrokerageEnvConfiguration, Configuration, InternalInputUserInput, \
    PathParameterUserInput
from copy import copy
from abc import ABC

_logged_messages = set()


class JsonModule(ABC):
    """The JsonModule class is the base class extended for all json modules."""

    def __init__(self, json_module_data: Dict[str, Any], module_type: str, platform: str) -> None:
        self._module_type: str = module_type
        self._platform: str = platform
        self._product_id: int = json_module_data["product-id"] if "product-id" in json_module_data else 0
        self._id: str = json_module_data["id"]
        self._display_name: str = json_module_data["display-id"]
        self._installs: bool = json_module_data["installs"] if ("installs" in json_module_data
                                                                and platform == MODULE_CLI_PLATFORM) else False
        self._lean_configs: List[Configuration] = []
        for config in json_module_data["configurations"]:
            self._lean_configs.append(Configuration.factory(config))
        self._lean_configs = self.sort_configs()
        self._is_module_installed: bool = False
        self._initial_cash_balance: LiveInitialStateInput = LiveInitialStateInput(
            json_module_data["live-cash-balance-state"]) \
            if "live-cash-balance-state" in json_module_data \
            else None
        self._initial_holdings: LiveInitialStateInput = LiveInitialStateInput(json_module_data["live-holdings-state"]) \
            if "live-holdings-state" in json_module_data \
            else False
        self._minimum_seat = json_module_data["minimum-seat"] if "minimum-seat" in json_module_data else None

    def get_id(self):
        return self._id

    def sort_configs(self) -> List[Configuration]:
        sorted_configs = []
        brokerage_configs = []
        for config in self._lean_configs:
            if isinstance(config, BrokerageEnvConfiguration):
                brokerage_configs.append(config)
            else:
                sorted_configs.append(config)
        return brokerage_configs + sorted_configs

    def get_name(self) -> str:
        """Returns the user-friendly name which users can identify this object by.

        :return: the user-friendly name to display to users
        """
        return self._display_name

    def _check_if_config_passes_filters(self, config: Configuration, all_for_platform_type: bool) -> bool:
        for condition in config._filter._conditions:
            if condition._dependent_config_id == MODULE_TYPE:
                target_value = self._module_type
            elif condition._dependent_config_id == MODULE_PLATFORM:
                target_value = self._platform
            else:
                if all_for_platform_type:
                    # skip, we want all configurations that match type and platform, for help
                    continue
                target_value = self.get_config_value_from_name(condition._dependent_config_id)
            if not target_value or not condition.check(target_value):
                return False
        return True

    def get_config_value_from_name(self, target_name: str) -> str:
        [idx] = [i for i in range(len(self._lean_configs))
                 if self._lean_configs[i]._id == target_name]
        return self._lean_configs[idx]._value

    def is_value_in_config(self, searched_value: str) -> bool:
        searched_value = searched_value.lower()
        for i in range(len(self._lean_configs)):
            value = self._lean_configs[i]._value
            if isinstance(value, str):
                value = value.lower()
            if isinstance(value, list):
                value = [x.lower() for x in value]

            if searched_value in value:
                return True
        return False

    def get_settings(self) -> Dict[str, str]:
        settings: Dict[str, str] = {"id": self._id}

        # we build these after the rest, because they might depend on their values
        for config in self._lean_configs:
            if type(config) is InternalInputUserInput:
                if config._is_conditional:
                    for option in config._value_options:
                        if option._condition.check(self.get_config_value_from_name(option._condition._dependent_config_id)):
                            config._value = option._value
                            break
                    if not config._value:
                        options_to_log = set([(opt._condition._dependent_config_id,
                                               self.get_config_value_from_name(opt._condition._dependent_config_id))
                                              for opt in config._value_options])
                        raise ValueError(
                            f'No condition matched among present options for "{config._id}". '
                            f'Please review ' +
                            ', '.join([f'"{x[0]}"' for x in options_to_log]) +
                            f' given value{"s" if len(options_to_log) > 1 else ""} ' +
                            ', '.join([f'"{x[1]}"' for x in options_to_log]))

        for configuration in self._lean_configs:
            if not self._check_if_config_passes_filters(configuration, all_for_platform_type=False):
                continue
            settings[configuration._id] = str(configuration._value).replace("\\", "/")

        return settings

    def get_all_input_configs(self, filters: List[Type[Configuration]] = []) -> List[Configuration]:
        return [copy(config) for config in self._lean_configs if config._is_required_from_user
                if type(config) not in filters
                and self._check_if_config_passes_filters(config, all_for_platform_type=True)]

    def convert_lean_key_to_variable(self, lean_key: str) -> str:
        """Replaces hyphens with underscore to follow python naming convention.

        :param lean_key: string that uses hyphnes as separator. Used in lean config
        """
        return lean_key.replace('-', '_')

    def convert_variable_to_lean_key(self, variable_key: str) -> str:
        """Replaces underscore with hyphens to follow lean config naming convention.

        :param variable_key: string that uses underscore as separator as per python convention.
        """
        return variable_key.replace('_', '-')

    def config_build(self,
                     lean_config: Dict[str, Any],
                     logger: Logger,
                     interactive: bool,
                     properties: Dict[str, Any] = {},
                     hide_input: bool = False,
                     environment_name: str = None) -> 'JsonModule':
        """Builds a new instance of this class, prompting the user for input when necessary.

        :param lean_config: the Lean configuration dict to read defaults from
        :param logger: the logger to use
        :param interactive: true if running in interactive mode
        :param properties: the properties that passed as options
        :param hide_input: whether to hide secrets inputs
        :param environment_name: the target environment name
        :return: self
        """
        logger.debug(f'Configuring {self._display_name}')

        # filter properties that were not passed as command line arguments,
        # so that we prompt the user for them only when they don't have a value in the Lean config
        context = get_current_context()
        user_provided_options = {k: v for k, v in properties.items()
                                 if context.get_parameter_source(k) == ParameterSource.COMMANDLINE}

        missing_options = []
        for configuration in self._lean_configs:
            if not self._check_if_config_passes_filters(configuration, all_for_platform_type=False):
                continue
            if not configuration._is_required_from_user:
                continue
            # Let's log messages for internal input configurations as well
            if configuration._log_message is not None:
                log_message = configuration._log_message.strip()
                if log_message and log_message not in _logged_messages:
                    logger.info(log_message)
                    # make sure we log these messages once, we could use the same module for different functionalities
                    _logged_messages.add(log_message)
            if type(configuration) is InternalInputUserInput:
                continue

            property_name = self.convert_lean_key_to_variable(configuration._id)
            # Only ask for user input if the config wasn't given as an option
            if property_name in user_provided_options and user_provided_options[property_name]:
                user_choice = user_provided_options[property_name]
                logger.debug(
                    f'JsonModule({self._display_name}): user provided \'{user_choice}\' for \'{property_name}\'')
            else:
                logger.debug(f'JsonModule({self._display_name}): Configuration not provided \'{configuration._id}\'')
                user_choice = self.get_default(lean_config, configuration._id, environment_name, logger)

                # There's no value in the lean config, let's use the module default value instead and prompt the user
                # NOTE: using "not" instead of "is None" because the default value can be false,
                #       in which case we still want to prompt the user.
                if not user_choice:
                    if interactive:
                        default_value = configuration._input_default
                        user_choice = configuration.ask_user_for_input(default_value, logger, hide_input=hide_input)

                        if not isinstance(configuration, BrokerageEnvConfiguration):
                            self._save_property({f"{configuration._id}": user_choice})
                    else:
                        if configuration._input_default != None and configuration._optional:
                            # if optional and we have a default input value and the user didn't provider it we use it
                            user_choice = configuration._input_default
                        else:
                            missing_options.append(f"--{configuration._id}")

            configuration._value = user_choice

        if len(missing_options) > 0:
            raise RuntimeError(f"""You are missing the following option{"s" if len(missing_options) > 1 else ""}: {', '
                               .join(missing_options)}""".strip())
        return self

    def get_paths_to_mount(self) -> Dict[str, str]:
        return {config._id: config._value
                for config in self._lean_configs
                if (isinstance(config, PathParameterUserInput)
                    and self._check_if_config_passes_filters(config, all_for_platform_type=False))}

    def ensure_module_installed(self, organization_id: str) -> None:
        if not self._is_module_installed and self._installs:
            container.logger.debug(f"JsonModule.ensure_module_installed(): installing module {self}: {self._product_id}")
            container.module_manager.install_module(
                self._product_id, organization_id)
            self._is_module_installed = True

    def get_default(self, lean_config: Dict[str, Any], key: str, environment_name: str, logger: Logger):
        user_choice = None
        if lean_config is not None:
            if (environment_name and "environments" in lean_config and environment_name in lean_config["environments"]
                    and key in lean_config["environments"][environment_name]):
                user_choice = lean_config["environments"][environment_name][key]
                logger.debug(f'JsonModule({self._display_name}): found \'{user_choice}\' for \'{key}\', in environment')
            elif key in lean_config:
                user_choice = lean_config[key]
                logger.debug(f'JsonModule({self._display_name}): found \'{user_choice}\' for \'{key}\'')
        return user_choice

    def __repr__(self):
        return self.get_name()

    def _save_property(self, settings: Dict[str, Any]):
        from lean.container import container
        container.lean_config_manager.set_properties(settings)


class LiveInitialStateInput(str, Enum):
    Required = "required"
    Optional = "optional"
    NotSupported = "not-supported"



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

reserved_names = ["CON", "PRN", "AUX", "NUL",
                    "COM1", "COM2", "COM3", "COM4", "COM5", "COM6", "COM7", "COM8", "COM9",
                    "LPT1", "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", "LPT9"]

forbidden_characters = ["\\", ":", "*", "?", '"', "<", ">", "|"]



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from pathlib import Path
from click import command, option

from lean.click import LeanCommand, PathParameter
from lean.container import container
from lean.models.encryption import ActionType

@command(cls=LeanCommand)
@option("--project", type=str, help="Name or id of the project to pull (all cloud projects if not specified)")
@option("--pull-bootcamp", is_flag=True, default=False, help="Pull Boot Camp projects (disabled by default)")
@option("--encrypt",
        is_flag=True, default=False,
        help="Pull your cloud files and encrypt them before saving on your local drive")
@option("--decrypt",
        is_flag=True, default=False,
        help="Pull your cloud files and decrypt them before saving on your local drive")
@option("--key",
              type=PathParameter(exists=True, file_okay=True, dir_okay=False),
              help="Path to the encryption key to use")
def pull(project: Optional[str], pull_bootcamp: bool, encrypt: Optional[bool], decrypt: Optional[bool], key: Optional[Path]) -> None:
    """Pull projects from QuantConnect to the local drive.

    This command overrides the content of local files with the content of their respective counterparts in the cloud.

    This command will not delete local files for which there is no counterpart in the cloud.
    """

    encryption_action = None

    from lean.components.util.encryption_helper import validate_user_inputs_for_cloud_push_pull_commands
    validate_user_inputs_for_cloud_push_pull_commands(encrypt, decrypt, key)
    
    if encrypt:
        encryption_action = ActionType.ENCRYPT
    if decrypt:
        encryption_action = ActionType.DECRYPT

    # Parse which projects need to be pulled
    project_id = None
    project_name = None
    if project is not None:
        try:
            project_id = int(project)
        except ValueError:
            # We treat it as a name rather than an id
            project_name = project

    api_client = container.api_client
    projects_to_pull = []
    all_projects = None

    organization_id = container.organization_manager.try_get_working_organization_id()

    if project_id is not None:
        projects_to_pull.append(api_client.projects.get(project_id, organization_id))
    else:
        all_projects = api_client.projects.get_all(organization_id)
        project_manager = container.project_manager
        projects_to_pull = project_manager.get_projects_by_name_or_id(all_projects, project_name)

    if project is None and not pull_bootcamp:
        projects_to_pull = [p for p in projects_to_pull if not p.name.startswith("Boot Camp/")]

    if key is not None and len(projects_to_pull) > 1:
        raise RuntimeError(f"Cannot encrypt or decrypt more than one project at a time.")

    # the encryption key info is available when reading the project individually from API
    projects_to_pull = [api_client.projects.get(project.projectId, project.organizationId) if project.encrypted == True else project for project in projects_to_pull]

    pull_manager = container.pull_manager
    pull_manager.pull_projects(projects_to_pull, all_projects, encryption_action, key)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from click import Choice, option, argument

from lean.click import LeanCommand
from lean.commands import lean
from lean.container import container
from lean.models.api import QCLanguage
from lean.models.errors import MoreInfoError
from lean.components.util.name_extraction import convert_to_class_name
from lean.components import reserved_names

DEFAULT_PYTHON_MAIN = '''
# region imports
from AlgorithmImports import *
# endregion

class $CLASS_NAME$(QCAlgorithm):

    def Initialize(self):
        # Locally Lean installs free sample data, to download more data please visit https://www.quantconnect.com/docs/v2/lean-cli/datasets/downloading-data
        self.SetStartDate(2013, 10, 7)  # Set Start Date
        self.SetEndDate(2013, 10, 11)  # Set End Date
        self.SetCash(100000)  # Set Strategy Cash
        self.AddEquity("SPY", Resolution.Minute)

    def OnData(self, data: Slice):
        """OnData event is the primary entry point for your algorithm. Each new data point will be pumped in here.
            Arguments:
                data: Slice object keyed by symbol containing the stock data
        """
        if not self.Portfolio.Invested:
            self.SetHoldings("SPY", 1)
            self.Debug("Purchased Stock")
'''.strip() + "\n"

LIBRARY_PYTHON_MAIN = '''
#region imports
from AlgorithmImports import *
#endregion


### Library classes are snippets of code/classes you can reuse between projects. They are
### added to projects on compile.
###
### To import this class use the following import with your values subbed in for the {} sections:
### from {libraryProjectName} import Library
###
### Example using your newly imported library from 'Library.py' like so:
###
### from $PROJECT_NAME$ import Library
### x = Library.Add(1,1)
### print(x)
###

def Add(a: int, b: int):
    return a + b
'''.strip() + "\n"

DEFAULT_PYTHON_NOTEBOOK = """
{
    "cells": [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": [
                "![QuantConnect Logo](https://cdn.quantconnect.com/web/i/icon.png)\\n",
                "<hr>"
            ]
        },
        {
            "cell_type": "code",
            "execution_count": null,
            "metadata": {},
            "outputs": [],
            "source": [
                "# QuantBook Analysis Tool \\n",
                "# For more information see [https://www.quantconnect.com/docs/v2/our-platform/research/getting-started]\\n",
                "qb = QuantBook()\\n",
                "spy = qb.AddEquity(\\"SPY\\")\\n",
                "# Locally Lean installs free sample data, to download more data please visit https://www.quantconnect.com/docs/v2/lean-cli/datasets/downloading-data \\n",
                "qb.SetStartDate(2013, 10, 11)\\n",
                "history = qb.History(qb.Securities.Keys, 360, Resolution.Daily)\\n",
                "\\n",
                "# Indicator Analysis\\n",
                "bbdf = qb.Indicator(BollingerBands(30, 2), spy.Symbol, 360, Resolution.Daily)\\n",
                "bbdf.drop('standarddeviation', axis=1).plot()"
            ]
        }
    ],
    "metadata": {
        "kernelspec": {
            "display_name": "Python 3",
            "language": "python",
            "name": "python3"
        }
    },
    "nbformat": 4,
    "nbformat_minor": 2
}
""".strip() + "\n"

DEFAULT_CSHARP_MAIN = """
#region imports
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Linq;
    using System.Globalization;
    using System.Drawing;
    using QuantConnect;
    using QuantConnect.Algorithm.Framework;
    using QuantConnect.Algorithm.Framework.Selection;
    using QuantConnect.Algorithm.Framework.Alphas;
    using QuantConnect.Algorithm.Framework.Portfolio;
    using QuantConnect.Algorithm.Framework.Portfolio.SignalExports;
    using QuantConnect.Algorithm.Framework.Execution;
    using QuantConnect.Algorithm.Framework.Risk;
    using QuantConnect.Api;
    using QuantConnect.Parameters;
    using QuantConnect.Benchmarks;
    using QuantConnect.Brokerages;
    using QuantConnect.Configuration;
    using QuantConnect.Util;
    using QuantConnect.Interfaces;
    using QuantConnect.Algorithm;
    using QuantConnect.Indicators;
    using QuantConnect.Data;
    using QuantConnect.Data.Auxiliary;
    using QuantConnect.Data.Consolidators;
    using QuantConnect.Data.Custom;
    using QuantConnect.Data.Custom.IconicTypes;
    using QuantConnect.DataSource;
    using QuantConnect.Data.Fundamental;
    using QuantConnect.Data.Market;
    using QuantConnect.Data.Shortable;
    using QuantConnect.Data.UniverseSelection;
    using QuantConnect.Notifications;
    using QuantConnect.Orders;
    using QuantConnect.Orders.Fees;
    using QuantConnect.Orders.Fills;
    using QuantConnect.Orders.OptionExercise;
    using QuantConnect.Orders.Slippage;
    using QuantConnect.Orders.TimeInForces;
    using QuantConnect.Python;
    using QuantConnect.Scheduling;
    using QuantConnect.Securities;
    using QuantConnect.Securities.Equity;
    using QuantConnect.Securities.Future;
    using QuantConnect.Securities.Option;
    using QuantConnect.Securities.Positions;
    using QuantConnect.Securities.Forex;
    using QuantConnect.Securities.Crypto;
    using QuantConnect.Securities.CryptoFuture;
    using QuantConnect.Securities.Interfaces;
    using QuantConnect.Securities.Volatility;
    using QuantConnect.Storage;
    using QuantConnect.Statistics;
    using QCAlgorithmFramework = QuantConnect.Algorithm.QCAlgorithm;
    using QCAlgorithmFrameworkBridge = QuantConnect.Algorithm.QCAlgorithm;
#endregion
namespace QuantConnect.Algorithm.CSharp
{
    public class $CLASS_NAME$ : QCAlgorithm
    {

        public override void Initialize()
        {
            // Locally Lean installs free sample data, to download more data please visit https://www.quantconnect.com/docs/v2/lean-cli/datasets/downloading-data
            SetStartDate(2013, 10, 7); // Set Start Date
            SetEndDate(2013, 10, 11); // Set Start Date
            SetCash(100000);             //Set Strategy Cash

            AddEquity("SPY", Resolution.Minute);

        }

        /// OnData event is the primary entry point for your algorithm. Each new data point will be pumped in here.
        /// Slice object keyed by symbol containing the stock data
        public override void OnData(Slice data)
        {
            if (!Portfolio.Invested)
            {
                SetHoldings("SPY", 1);
                Debug("Purchased Stock");
            }
        }

    }
}
""".strip() + "\n"

LIBRARY_CSHARP_MAIN = """
#region imports
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Linq;
    using System.Globalization;
    using System.Drawing;
    using QuantConnect;
    using QuantConnect.Algorithm.Framework;
    using QuantConnect.Algorithm.Framework.Selection;
    using QuantConnect.Algorithm.Framework.Alphas;
    using QuantConnect.Algorithm.Framework.Portfolio;
    using QuantConnect.Algorithm.Framework.Portfolio.SignalExports;
    using QuantConnect.Algorithm.Framework.Execution;
    using QuantConnect.Algorithm.Framework.Risk;
    using QuantConnect.Api;
    using QuantConnect.Parameters;
    using QuantConnect.Benchmarks;
    using QuantConnect.Brokerages;
    using QuantConnect.Configuration;
    using QuantConnect.Util;
    using QuantConnect.Interfaces;
    using QuantConnect.Algorithm;
    using QuantConnect.Indicators;
    using QuantConnect.Data;
    using QuantConnect.Data.Auxiliary;
    using QuantConnect.Data.Consolidators;
    using QuantConnect.Data.Custom;
    using QuantConnect.Data.Custom.IconicTypes;
    using QuantConnect.DataSource;
    using QuantConnect.Data.Fundamental;
    using QuantConnect.Data.Market;
    using QuantConnect.Data.Shortable;
    using QuantConnect.Data.UniverseSelection;
    using QuantConnect.Notifications;
    using QuantConnect.Orders;
    using QuantConnect.Orders.Fees;
    using QuantConnect.Orders.Fills;
    using QuantConnect.Orders.OptionExercise;
    using QuantConnect.Orders.Slippage;
    using QuantConnect.Orders.TimeInForces;
    using QuantConnect.Python;
    using QuantConnect.Scheduling;
    using QuantConnect.Securities;
    using QuantConnect.Securities.Equity;
    using QuantConnect.Securities.Future;
    using QuantConnect.Securities.Option;
    using QuantConnect.Securities.Positions;
    using QuantConnect.Securities.Forex;
    using QuantConnect.Securities.Crypto;
    using QuantConnect.Securities.CryptoFuture;
    using QuantConnect.Securities.Interfaces;
    using QuantConnect.Securities.Volatility;
    using QuantConnect.Storage;
    using QuantConnect.Statistics;
    using QCAlgorithmFramework = QuantConnect.Algorithm.QCAlgorithm;
    using QCAlgorithmFrameworkBridge = QuantConnect.Algorithm.QCAlgorithm;
#endregion


namespace QuantConnect
{
    /// <summary>
    /// Template Library Class
    ///
    /// Library classes are snippets of code you can reuse between projects. They are added to projects on compile. This can be useful for reusing
    /// indicators, math components, risk modules etc. If you use a custom namespace make sure you add the correct using statement to the
    /// algorithm.
    /// </summary>
    public static class $CLASS_NAME$
    {
        /*
         * To use this library; add its namespace at the top of the page:
         * using QuantConnect
         *
         * Then use the static class:
         * var x = $CLASS_NAME$.Add(1, 5);
         * Console.Out.WriteLine(x);
         */

        public static int Add(int a, int b)
        {
            return a + b;
        }

        public static int Subtract(int a, int b)
        {
            return a - b;
        }

        public static int Divide(int a, int b){
            return a / b;
        }
    }
}
""".strip() + "\n"

DEFAULT_CSHARP_NOTEBOOK = """
{
    "cells": [
        {
            "cell_type": "markdown",
            "metadata": {},
            "source": [
                "![QuantConnect Logo](https://cdn.quantconnect.com/web/i/icon.png)\\n",
                "<hr>"
            ]
        },
        {
            "cell_type": "code",
            "execution_count": null,
            "metadata": {},
            "outputs": [],
            "source": [
                "// We need to load assemblies at the start in their own cell\\n",
                "#load \\"../Initialize.csx\\""
            ]
        },
        {
            "cell_type": "code",
            "execution_count": null,
            "metadata": {},
            "outputs": [],
            "source": [
                "// QuantBook C# Research Environment\\n",
                "// For more information see https://www.quantconnect.com/docs/v2/our-platform/research/getting-started\\n",
                "#load \\"../QuantConnect.csx\\"\\n",
                "\\n",
                "using QuantConnect;\\n",
                "using QuantConnect.Data;\\n",
                "using QuantConnect.Research;\\n",
                "using QuantConnect.Algorithm;\\n",
                "\\n",
                "var qb = new QuantBook();\\n",
                "// Locally Lean installs free sample data, to download more data please visit https://www.quantconnect.com/docs/v2/lean-cli/datasets/downloading-data \\n",
                "qb.SetStartDate(2013, 10, 11);\\n",
                "var spy = qb.AddEquity(\\"SPY\\");\\n",
                "var history = qb.History(qb.Securities.Keys, 360, Resolution.Daily);"
            ]
        },
        {
            "cell_type": "code",
            "execution_count": null,
            "metadata": {},
            "outputs": [],
            "source": [
                "foreach (var slice in history.Take(5)) {\\n",
                "    Console.WriteLine(slice.Bars[spy.Symbol].ToString());\\n",
                "}"
            ]
        },
        {
            "cell_type": "code",
            "execution_count": null,
            "metadata": {},
            "outputs": [],
            "source": []
        }
    ],
    "metadata": {
        "kernelspec": {
            "display_name": ".NET (C#)",
            "language": "C#",
            "name": ".net-csharp"
        },
        "language_info": {
            "file_extension": ".cs",
            "mimetype": "text/x-csharp",
            "name": "C#",
            "pygments_lexer": "csharp",
            "version": "9.0"
        }
    },
    "nbformat": 4,
    "nbformat_minor": 2
}
""".strip() + "\n"

def _not_identifier_char(text):
    problematic_char = text[-1]
    for i in range(1, len(text)):
        substring = text[:i]
        if not substring.isidentifier():
            problematic_char = substring[-1]
            break
    return problematic_char

@lean.command(cls=LeanCommand, name="project-create", aliases=["create-project"])
@argument("name", type=str)
@option("--language", "-l",
              type=Choice(container.cli_config_manager.default_language.allowed_values, case_sensitive=False),
              help="The language of the project to create")
def create_project(name: str, language: str) -> None:
    """Create a new project containing starter code.

    If NAME is a path containing subdirectories those will be created automatically.

    The default language can be set using `lean config set default-language python/csharp`.
    """
    cli_config_manager = container.cli_config_manager

    language = language if language is not None else cli_config_manager.default_language.get_value()
    if language is None:
        raise MoreInfoError(
            "Please specify a language with --language or set the default language using `lean config set default-language python/csharp`",
            "https://www.lean.io/docs/v2/lean-cli/projects/project-management")

    full_path = Path.cwd() / name
    try:
        cli_root_dir = container.lean_config_manager.get_cli_root_directory()
        relative_path = full_path.relative_to(cli_root_dir).as_posix()
    except MoreInfoError:
        relative_path = name

    if not container.path_manager.is_cli_path_valid(full_path) or not container.path_manager.is_name_valid(relative_path):
        raise MoreInfoError(f"Invalid project name. Can only contain letters, numbers & spaces. Can not start with empty char ' ' or be a reserved name [ {', '.join(reserved_names)} ]",
                         "https://www.lean.io/docs/v2/lean-cli/key-concepts/troubleshooting#02-Common-Errors")

    is_library_project = False
    try:
        library_dir = container.lean_config_manager.get_cli_root_directory() / "Library"
        is_library_project = library_dir in full_path.parents
        if is_library_project:
            # Make sure we always use the same casing 'Library' for the library directory
            full_path = library_dir / full_path.relative_to(library_dir)
    except:
        # get_cli_root_directory() raises an error if there is no such directory
        pass

    id_name = full_path.name
    if is_library_project and language == "python" and not id_name.isidentifier():
        problematic_char = _not_identifier_char(id_name)
        raise RuntimeError(
            f"""'{id_name}' is not a valid Python identifier, which is required for Python library projects to be importable.
Please remove the character '{problematic_char}' and retry""")

    if full_path.exists():
        raise RuntimeError(f"A project named '{name}' already exists, please choose a different name")
    else:
        project_manager = container.project_manager
        project_manager.create_new_project(full_path, QCLanguage.Python if language == "python" else QCLanguage.CSharp)

    class_name = convert_to_class_name(full_path)

    if language == "python":
        main_name = "main.py" if not is_library_project else "Library.py"
        research_name = "research.ipynb"
        main_content = DEFAULT_PYTHON_MAIN if not is_library_project else LIBRARY_PYTHON_MAIN
    else:
        main_name = "Main.cs" if not is_library_project else "Library.cs"
        research_name = "Research.ipynb"
        main_content = DEFAULT_CSHARP_MAIN if not is_library_project else LIBRARY_CSHARP_MAIN

    with (full_path / main_name).open("w+", encoding="utf-8") as file:
        file.write(main_content.replace("$CLASS_NAME$", class_name).replace("$PROJECT_NAME$", full_path.name))

    if not is_library_project:
        with (full_path / research_name).open("w+", encoding="utf-8") as file:
            file.write(DEFAULT_PYTHON_NOTEBOOK if language == "python" else DEFAULT_CSHARP_NOTEBOOK)

    if language == "csharp":
        project_manager = container.project_manager
        project_csproj_file = project_manager.get_csproj_file_path(full_path)
        original_csproj_content = project_csproj_file.read_text(encoding="utf-8")
        project_manager.try_restore_csharp_project(project_csproj_file, original_csproj_content, False)

    logger = container.logger
    logger.info(f"Successfully created {'Python' if language == 'python' else 'C#'} project '{name}'")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional, Tuple
from click import command, argument, option, Choice
from lean.click import LeanCommand, PathParameter
from lean.components.docker.lean_runner import LeanRunner
from lean.constants import DEFAULT_RESEARCH_IMAGE, LEAN_ROOT_PATH
from lean.container import container
from lean.models.cli import cli_data_downloaders
from lean.components.util.name_extraction import convert_to_class_name
from lean.components.util.json_modules_handler import non_interactive_config_build_for_name
from lean.models.click_options import options_from_json, get_configs_for_options

def _check_docker_output(chunk: str, port: int) -> None:
    """Checks the output of the Docker container and opens the browser if Jupyter Lab has started.

    :param chunk: the output chunk
    :param port: the port Jupyter Lab will be running on
    """
    from webbrowser import open
    if "is running at:" in chunk:
        open(f"http://localhost:{port}/")


@command(cls=LeanCommand, requires_lean_config=True, requires_docker=True)
@argument("project", type=PathParameter(exists=True, file_okay=False, dir_okay=True))
@option("--port", type=int, default=8888, help="The port to run Jupyter Lab on (defaults to 8888)")
@option("--data-provider-historical",
              type=Choice([dp.get_name() for dp in cli_data_downloaders], case_sensitive=False),
              default="Local",
              help="Update the Lean configuration file to retrieve data from the given historical provider")
@options_from_json(get_configs_for_options("research"))
@option("--download-data",
              is_flag=True,
              default=False,
              help=f"Update the Lean configuration file to download data from the QuantConnect API, alias for --data-provider-historical QuantConnect")
@option("--data-purchase-limit",
              type=int,
              help="The maximum amount of QCC to spend on downloading data during the research session when using QuantConnect as historical data provider")
@option("--detach", "-d",
              is_flag=True,
              default=False,
              help="Run Jupyter Lab in a detached Docker container and return immediately")
@option("--no-open",
              is_flag=True,
              default=False,
              help="Don't open the Jupyter Lab environment in the browser after starting it")
@option("--image", type=str, help=f"The LEAN research image to use (defaults to {DEFAULT_RESEARCH_IMAGE})")
@option("--update",
              is_flag=True,
              default=False,
              help="Pull the LEAN research image before starting the research environment")
@option("--extra-config",
              type=(str, str),
              multiple=True,
              hidden=True)
@option("--extra-docker-config",
              type=str,
              default="{}",
              help="Extra docker configuration as a JSON string. "
                   "For more information https://docker-py.readthedocs.io/en/stable/containers.html")
@option("--no-update",
              is_flag=True,
              default=False,
              help="Use the local LEAN research image instead of pulling the latest version")
def research(project: Path,
             port: int,
             data_provider_historical: Optional[str],
             download_data: bool,
             data_purchase_limit: Optional[int],
             detach: bool,
             no_open: bool,
             image: Optional[str],
             update: bool,
             extra_config: Optional[Tuple[str, str]],
             extra_docker_config: Optional[str],
             no_update: bool,
             **kwargs) -> None:
    """Run a Jupyter Lab environment locally using Docker.

    By default the official LEAN research image is used.
    You can override this using the --image option.
    Alternatively you can set the default research image using `lean config set research-image <image>`.
    """
    from docker.types import Mount
    from docker.errors import APIError
    from json import loads

    logger = container.logger

    project_manager = container.project_manager
    algorithm_file = project_manager.find_algorithm_file(project)
    algorithm_name = convert_to_class_name(project)

    environment_name = "backtesting"
    lean_config_manager = container.lean_config_manager
    lean_config = lean_config_manager.get_complete_lean_config(environment_name, algorithm_file, None)
    lean_config["composer-dll-directory"] = LEAN_ROOT_PATH
    lean_config["research-object-store-name"] = algorithm_name

    if download_data:
        data_provider_historical = "QuantConnect"

    paths_to_mount = None

    if data_provider_historical is not None:
        organization_id = container.organization_manager.try_get_working_organization_id()
        data_provider = non_interactive_config_build_for_name(lean_config, data_provider_historical,
                                                              cli_data_downloaders, kwargs, logger, environment_name)
        data_provider.ensure_module_installed(organization_id)
        container.lean_config_manager.set_properties(data_provider.get_settings())
        paths_to_mount = data_provider.get_paths_to_mount()
    lean_config_manager.configure_data_purchase_limit(lean_config, data_purchase_limit)

    lean_runner = container.lean_runner
    temp_manager = container.temp_manager

    # Set extra config
    for key, value in extra_config:
        lean_config[key] = value

    project_config_manager = container.project_config_manager
    cli_config_manager = container.cli_config_manager

    project_config = project_config_manager.get_project_config(algorithm_file.parent)
    research_image = cli_config_manager.get_research_image(image or project_config.get("research-image", None))

    container.update_manager.pull_docker_image_if_necessary(research_image, update, no_update)

    if str(research_image) != DEFAULT_RESEARCH_IMAGE:
        logger.warn(f'A custom research image: "{research_image}" is being used!')

    run_options = lean_runner.get_basic_docker_config(lean_config,
                                                      algorithm_file,
                                                      temp_manager.create_temporary_directory(),
                                                      None,
                                                      False,
                                                      detach,
                                                      research_image,
                                                      paths_to_mount)

    # Mount the config in the notebooks directory as well
    local_config_path = next(m["Source"] for m in run_options["mounts"] if m["Target"].endswith("config.json"))
    run_options["mounts"].append(Mount(target=f"{LEAN_ROOT_PATH}/Notebooks/config.json",
                                       source=str(local_config_path),
                                       type="bind",
                                       read_only=True))

    # Jupyter Lab runs on port 8888, we expose it to the local port specified by the user
    run_options["ports"]["8888"] = str(port)

    # Open the browser as soon as Jupyter Lab has started
    if detach or not no_open:
        run_options["on_output"] = lambda chunk: _check_docker_output(chunk, port)

    # Make Ctrl+C stop Jupyter Lab immediately
    run_options["stop_signal"] = "SIGKILL"

    # Mount the project to the notebooks directory
    run_options["volumes"][str(project)] = {
        "bind": f"{LEAN_ROOT_PATH}/Notebooks",
        "mode": "rw"
    }

    # Allow notebooks to be embedded in iframes
    run_options["commands"].append("mkdir -p ~/.jupyter")
    run_options["commands"].append(
        'echo "c.ServerApp.disable_check_xsrf = True\nc.ServerApp.tornado_settings = {\'headers\': {\'Content-Security-Policy\': \'frame-ancestors self *\'}}" > ~/.jupyter/jupyter_server_config.py')

    # Hide headers in notebooks
    run_options["commands"].append("mkdir -p ~/.ipython/profile_default/static/custom")
    run_options["commands"].append(
        'echo "#header-container { display: none !important; }" > ~/.ipython/profile_default/static/custom/custom.css')

    # Run the script that starts Jupyter Lab when all set up has been done
    run_options["commands"].append("./start.sh")

    # Add known additional run options from the extra docker config
    LeanRunner.parse_extra_docker_config(run_options, loads(extra_docker_config))

    try:
        container.docker_manager.run_image(research_image, **run_options)
    except APIError as error:
        msg = error.explanation
        if isinstance(msg, str) and any(m in msg.lower() for m in [
            "port is already allocated",
            "ports are not available"
            "an attempt was made to access a socket in a way forbidden by its access permissions"
        ]):
            raise RuntimeError(f"Port {port} is already in use, please specify a different port using --port <number>")
        raise error

    if detach:
        temp_manager.delete_temporary_directories_when_done = False

        relative_project_dir = algorithm_file.parent.relative_to(lean_config_manager.get_cli_root_directory())

        logger.info(
            f"Successfully started Jupyter Lab environment for '{relative_project_dir}' in the '{run_options['name']}' container")
        logger.info("You can use Docker's own commands to manage the detached container")



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Dict
from click import option, Choice
from lean.click import PathParameter
from lean.models.cli import cli_brokerages, cli_data_downloaders, cli_data_queue_handlers
from lean.models.cloud import cloud_brokerages, cloud_data_queue_handlers
from lean.models.configuration import Configuration, InfoConfiguration, InternalInputUserInput


def get_configs_for_options(env: str) -> List[Configuration]:
    if env == "live-cloud":
        brokerage = cloud_brokerages + cloud_data_queue_handlers
    elif env == "live-cli":
        brokerage = cli_brokerages + cli_data_queue_handlers + cli_data_downloaders
    elif env == "backtest":
        brokerage = cli_data_downloaders
    elif env == "research":
        brokerage = cli_data_downloaders
    else:
        raise ValueError("Acceptable values for 'env' are: 'live-cloud', 'live-cli', 'backtest', 'research'")

    run_options: Dict[str, Configuration] = {}
    config_with_module_id: Dict[str, str] = {}
    for module in brokerage:
        for config in module.get_all_input_configs([InternalInputUserInput, InfoConfiguration]):
            if config._id in run_options:
                if config._id in config_with_module_id and config_with_module_id[config._id] == module._id:
                    continue
                else:
                    raise ValueError(f'Options names should be unique. Duplicate key present: {config._id}')
            run_options[config._id] = config
            config_with_module_id[config._id] = module._id
    return list(run_options.values())


def get_click_option_type(configuration: Configuration):
    # get type should be a method of configurations class itself.
    # TODO: handle input can inherit type prompt.
    if configuration._config_type == "internal-input":
        return str
    if configuration._input_method == "confirm":
        return bool
    elif configuration._input_method == "choice":
        return Choice(configuration._choices, case_sensitive=False)
    elif configuration._input_method == "prompt":
        return configuration.get_input_type()
    elif configuration._input_method == "prompt-password":
        return str
    elif configuration._input_method == "path-parameter":
        return PathParameter(exists=False, file_okay=True, dir_okay=False)


def get_attribute_type(configuration: Configuration):
    # get type should be a method of configurations class itself.
    # TODO: handle input can inherit type prompt.
    if configuration._config_type == "internal-input":
        return str
    if configuration._input_method == "confirm":
        return bool
    elif configuration._input_method == "choice":
        return str
    elif configuration._input_method == "prompt":
        return configuration.get_input_type()
    elif configuration._input_method == "prompt-password":
        return str
    elif configuration._input_method == "path-parameter":
        return str


def get_options_attributes(configuration: Configuration, default_lean_config_key=None):
    options_attributes = {
        "type": get_click_option_type(configuration),
        "help": configuration._help
    }
    return options_attributes


def get_default_key(configuration: Configuration):
    return configuration._id


def options_from_json(configurations: List[Configuration]):

    def decorator(f):
        for configuration in reversed(configurations):
            long = configuration._id
            name = str(configuration._id).replace('-', '_')
            param_decls = (
                '--' + long,
                name)
            attrs = get_options_attributes(
                configuration, get_default_key(configuration))
            option(*param_decls, **attrs)(f)
        return f
    return decorator



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import List, Dict, Optional

from lean.components.api.api_client import APIClient
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.util.logger import Logger
from lean.components.util.organization_manager import OrganizationManager
from lean.components.util.project_manager import ProjectManager
from lean.models.api import QCLanguage, QCProject
from lean.models.utils import LeanLibraryReference
from lean.models.encryption import ActionType

class PushManager:
    """The PushManager class is responsible for synchronizing local projects to the cloud."""

    def __init__(self,
                 logger: Logger,
                 api_client: APIClient,
                 project_manager: ProjectManager,
                 project_config_manager: ProjectConfigManager,
                 organization_manager: OrganizationManager) -> None:
        """Creates a new PushManager instance.

        :param logger: the logger to use when printing messages
        :param api_client: the APIClient instance to use when communicating with the cloud
        :param project_manager: the ProjectManager to use when looking for certain projects
        :param project_config_manager: the ProjectConfigManager instance to use
        """
        self._logger = logger
        self._api_client = api_client
        self._project_manager = project_manager
        self._project_config_manager = project_config_manager
        self._organization_manager = organization_manager
        self._cloud_projects = []

    def push_project(self, project: Path, encryption_action: Optional[ActionType]=None, encryption_key: Optional[Path]=None) -> None:
        """Pushes the given project from the local drive to the cloud.

        It will also push every library referenced by the project and add or remove references.

        :param project: path to the directory containing the local project that needs to be pushed
        """
        libraries = self._project_manager.get_project_libraries(project)
        self.push_projects([project], libraries, encryption_action, encryption_key)

    def push_projects(self, projects_to_push: List[Path], associated_libraries_to_push: Optional[List[Path]]=[], encryption_action: Optional[ActionType]=None, encryption_key: Optional[Path]=None) -> None:
        """Pushes the given projects from the local drive to the cloud.

        It will also push every library referenced by each project and add or remove references.

        :param projects_to_push: a list of directories containing the local projects that need to be pushed
        """
        all_projects_to_push = associated_libraries_to_push + projects_to_push

        if len(all_projects_to_push) == 0:
            return

        organization_id = self._organization_manager.try_get_working_organization_id()

        for index, path in enumerate(all_projects_to_push, start=1):

            # Check if it's an associated library to push, we don't encrypt those.
            encryption_action_value = encryption_action if path not in associated_libraries_to_push else None
            encryption_key_value = encryption_key if path not in associated_libraries_to_push else None
            relative_path = path.relative_to(Path.cwd())
            try:
                self._logger.info(f"[{index}/{len(all_projects_to_push)}] Pushing '{relative_path}'")
                self._push_project(path, organization_id, encryption_action_value, encryption_key_value)
            except Exception as ex:
                from traceback import format_exc
                self._logger.debug(format_exc().strip())
                self._logger.warn(f"Cannot push '{relative_path}': {ex}")

    def _get_local_libraries_cloud_ids(self, project_dir: Path) -> List[int]:
        project_config = self._project_config_manager.get_project_config(project_dir)

        libraries_in_config = project_config.get("libraries", [])
        library_paths = [LeanLibraryReference(**library).path.expanduser().resolve() for library in libraries_in_config]

        local_libraries_cloud_ids = [int(self._project_config_manager.get_project_config(path).get("cloud-id", None))
                                     for path in library_paths]

        return local_libraries_cloud_ids

    def _push_project(self, project_path: Path, organization_id: str, encryption_action: Optional[ActionType], encryption_key: Optional[Path], suggested_rename_path: Path = None) -> None:
        """Pushes a single local project to the cloud.

        Raises an error with a descriptive message if the project cannot be pushed.

        :param project_path: the local project to push
        :param organization_id: the id of the organization to push the project to
        :param suggested_rename_path: the path to move the project to.
        """

        project_name = project_path.relative_to(Path.cwd()).as_posix()

        potential_new_name = project_name
        if suggested_rename_path and suggested_rename_path != project_path:
            potential_new_name = suggested_rename_path.relative_to(Path.cwd()).as_posix()


        project_config = self._project_config_manager.get_project_config(project_path)
        cloud_id = project_config.get("cloud-id")
        local_encryption_state = project_config.get("encrypted", False)
        local_encryption_key = project_config.get("encryption-key-path", None)
        if local_encryption_key is not None:
            local_encryption_key = Path(local_encryption_key)

        # check if project name is valid or if rename is required
        if cloud_id is not None:
            expected_correct_project_path = self._project_manager.get_local_project_path(potential_new_name, cloud_id)
        else:
            local_id = self._project_config_manager.get_local_id(project_path)
            expected_correct_project_path = self._project_manager.get_local_project_path(potential_new_name, None, local_id)
        if expected_correct_project_path != project_path:
            # rename project
            valid_project_name = expected_correct_project_path.relative_to(Path.cwd()).as_posix()
            self._logger.info(f"Renaming '{project_name}' to '{valid_project_name}'")
            self._project_manager.rename_project_and_contents(project_path, expected_correct_project_path)
            project_path = expected_correct_project_path
            project_name = valid_project_name
            project_config = self._project_config_manager.get_project_config(project_path)

        # Find the cloud project to push the files to
        if cloud_id is not None:
            # Project has cloud id which matches cloud project, update cloud project
            cloud_project = self._get_cloud_project(cloud_id, organization_id)
        else:
            # Project has invalid cloud id or no cloud id at all, create new cloud project
            cloud_project = self._api_client.projects.create(project_name,
                                                             QCLanguage[project_config.get("algorithm-language")],
                                                             organization_id)
            project_config.set("cloud-id", cloud_project.projectId)
            project_config.set("organization-id", cloud_project.organizationId)

            if cloud_project.name != project_name:
                # cloud project name was changed. Repeat steps to validate the new name locally.
                self._logger.info(f"Received new name '{cloud_project.name}' for project '{project_name}' from QuantConnect.com")
                self._push_project(project_path, organization_id, encryption_action, encryption_key, Path.cwd() / cloud_project.name)
                return

            self._cloud_projects.append(cloud_project)
            organization_message_part = f" in organization '{organization_id}'" if organization_id is not None else ""
            self._logger.info(f"Successfully created cloud project '{cloud_project.name}'{organization_message_part}")

        # Handle mismatch cases
        from lean.components.util.encryption_helper import validate_key_and_encryption_state_for_cloud_project
        validate_key_and_encryption_state_for_cloud_project(cloud_project, local_encryption_state, encryption_key, local_encryption_key, self._logger)
        if local_encryption_state == True and encryption_key is None and local_encryption_key is not None:
            encryption_key = local_encryption_key
            encryption_action = ActionType.ENCRYPT if local_encryption_state else ActionType.DECRYPT
        # Finalize pushing by updating locally modified metadata, files and libraries
        self._push_metadata(project_path, cloud_project, encryption_action, encryption_key)

    def _get_files(self, project: Path, encryption_action: Optional[ActionType], encryption_key: Optional[Path]) -> List[Dict[str, str]]:
        """Pushes the files of a local project to the cloud.

        :param project: the local project to push the files of
        """
        paths = self._project_manager.get_source_files(project)
        if encryption_key:
            from lean.components.util.encryption_helper import get_appropriate_files_from_local_project
            organization_id = self._organization_manager.try_get_working_organization_id()
            data = get_appropriate_files_from_local_project(project, paths, encryption_key, self._project_config_manager, organization_id, encryption_action)
            files = [
            {
                'name': path.relative_to(project).as_posix(),
                'content': encrypted_content
            }
            for path, encrypted_content in zip(paths, data)
        ]
        else:
            files = [
                {
                    'name': path.relative_to(project).as_posix(),
                    'content': path.read_text(encoding="utf-8")
                }
                for path in paths
            ]

        return files

    def _push_metadata(self, project: Path, cloud_project: QCProject, encryption_action: Optional[ActionType], encryption_key: Optional[Path]) -> None:
        """Pushes local project description and parameters to the cloud.

        Does nothing if the cloud is already up-to-date.

        :param project: the local project to push the parameters of
        :param cloud_project: the cloud project to push the parameters to
        """
        project_config = self._project_config_manager.get_project_config(project)

        local_description = project_config.get("description", "")
        cloud_description = cloud_project.description

        local_parameters = project_config.get("parameters", {})
        cloud_parameters = {parameter.key: parameter.value for parameter in cloud_project.parameters}

        # Use latest (-1) by default
        local_lean_version = int(project_config.get("lean-engine", "-1"))
        cloud_lean_version = cloud_project.leanVersionId

        local_lean_venv = project_config.get("python-venv", None)
        cloud_lean_venv = cloud_project.leanEnvironment

        update_args = {}

        expected_correct_project_name = project.relative_to(Path.cwd()).as_posix()

        # update project name in cloud in case it was incorrect and renamed locally otherwise update the same name
        update_args["name"] = expected_correct_project_name
        if cloud_project.name != expected_correct_project_name:
            self._logger.info(f"Renaming project in cloud from '{cloud_project.name}' to '{expected_correct_project_name}'")

        if local_description != cloud_description:
            update_args["description"] = local_description

        if local_parameters != cloud_parameters:
            update_args["parameters"] = local_parameters

        if (local_lean_version != cloud_lean_version and
           (local_lean_version != -1 or not cloud_project.leanPinnedToMaster)):
            update_args["lean_engine"] = local_lean_version

        # Initially, python-venv is not defined in the config and the default one will be used.
        # After it is changed, in order to use the default one again, it must not be removed from the config,
        # but it should be set to the default env id explicitly instead.
        if local_lean_venv is not None and local_lean_venv != cloud_lean_venv:
            update_args["python_venv"] = local_lean_venv

        update_args["files"] = self._get_files(project, encryption_action, encryption_key)
        update_args["libraries"] = self._get_local_libraries_cloud_ids(project)

        # default value
        update_args["encryption_key"] = ''
        if (encryption_action is not None and encryption_key is not None):
            # lets check if the given key is registered with the cloud
            from lean.components.util.encryption_helper import validate_encryption_key_registered_with_cloud, get_project_key_hash
            validate_encryption_key_registered_with_cloud(encryption_key, self._organization_manager, self._api_client)

            if encryption_action == ActionType.ENCRYPT:
                encryption_key_id = get_project_key_hash(encryption_key)
                update_args["encryption_key"] = encryption_key_id

        if update_args != {}:
            self._api_client.projects.update(cloud_project.projectId, **update_args)

            if "encryption_key" in update_args:
                del update_args["encryption_key"]
            updated_keys = list(update_args)
            if len(updated_keys) == 1:
                updated_keys_str = updated_keys[0]
            elif len(updated_keys) == 2:
                updated_keys_str = " and ".join(updated_keys)
            else:
                updated_keys_str = ", ".join(updated_keys[:-1]) + f", and {updated_keys[-1]}"
            self._logger.info(f"Successfully updated {updated_keys_str} for '{cloud_project.name}'")

    def _get_cloud_project(self, project_id: int, organization_id: str) -> QCProject:
        project = next(iter(p for p in self._cloud_projects if p.projectId == project_id), None)
        if project is None:
            project = self._api_client.projects.get(project_id, organization_id)
            self._cloud_projects.append(project)

        return project



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lean.components.config.lean_config_manager import LeanConfigManager

def open_storage_directory_in_explorer(lean_config_manager: LeanConfigManager):
    """Opens the storage directory in the file explorer."""
    global_storage_directory_path = lean_config_manager.get_cli_root_directory() / "storage"
    if not global_storage_directory_path.exists():
        global_storage_directory_path.mkdir(parents=True)
    open_file_explorer(str(global_storage_directory_path))
    
def open_file_explorer(directory_path: str):
    """Opens the given directory in the file explorer."""
    from webbrowser import open
    open('file:///' + directory_path)


# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union, Tuple
from lean.components import reserved_names
from lean.components.config.cli_config_manager import CLIConfigManager
from lean.components.config.lean_config_manager import LeanConfigManager
from lean.components.config.project_config_manager import ProjectConfigManager
from lean.components.docker.docker_manager import DockerManager
from lean.components.util.logger import Logger
from lean.components.util.path_manager import PathManager
from lean.components.util.platform_manager import PlatformManager
from lean.components.util.xml_manager import XMLManager
from lean.constants import PROJECT_CONFIG_FILE_NAME, DEFAULT_LEAN_DOTNET_FRAMEWORK
from lean.models.api import QCLanguage, QCProject, QCProjectLibrary
from lean.models.utils import LeanLibraryReference


class ProjectManager:
    """The ProjectManager class provides utilities for handling a single project."""

    def __init__(self,
                 logger: Logger,
                 project_config_manager: ProjectConfigManager,
                 lean_config_manager: LeanConfigManager,
                 path_manager: PathManager,
                 xml_manager: XMLManager,
                 platform_manager: PlatformManager,
                 cli_config_manager: CLIConfigManager,
                 docker_manager: DockerManager) -> None:
        """Creates a new ProjectManager instance.

        :param logger: the logger to use to log messages with
        :param project_config_manager: the ProjectConfigManager to use when creating new projects
        :param lean_config_manager: the LeanConfigManager to get the CLI root directory from
        :param path_manager: the path manager to use to handle library paths
        :param xml_manager: the XMLManager to use when working with XML
        :param platform_manager: the PlatformManager used when checking which operating system is in use
        """
        self._logger = logger
        self._project_config_manager = project_config_manager
        self._lean_config_manager = lean_config_manager
        self._path_manager = path_manager
        self._xml_manager = xml_manager
        self._platform_manager = platform_manager
        self._cli_config_manager = cli_config_manager
        self._docker_manager = docker_manager

    def find_algorithm_file(self, input: Path) -> Path:
        """Returns the path to the file containing the algorithm.

        Raises an error if the algorithm file cannot be found.

        :param input: the path to the algorithm or the path to the project
        :return: the path to the file containing the algorithm
        """
        if input.is_file():
            return input

        for file_name in ["main.py", "Main.cs"]:
            target_file = input / file_name
            if target_file.exists():
                return target_file

        raise ValueError("The specified project does not contain a main.py or Main.cs file")

    def get_project_by_id(self, local_id: int) -> Path:
        """Finds a project by its local id.

        Raises an error if a project with the given local id cannot be found.

        :param local_id: the local id of the project
        :return: the path to the directory containing the project with the given local id
        """
        directories = [self._lean_config_manager.get_cli_root_directory()]
        while len(directories) > 0:
            directory = directories.pop(0)

            config_file = directory / PROJECT_CONFIG_FILE_NAME
            if config_file.is_file():
                if self._project_config_manager.get_local_id(directory) == local_id:
                    return directory
            else:
                directories.extend(d for d in directory.iterdir() if d.is_dir())

        raise RuntimeError(f"Project with local id '{local_id}' does not exist")

    def try_get_project_path_by_cloud_id(self, cloud_id: int) -> Path:
        """Finds a project by its cloud id.

        Raises an error if a project with the given cloud id cannot be found.

        :param cloud_id: the cloud id of the project
        :return: the path to the directory containing the project with the given cloud id
        """
        directories = [self._lean_config_manager.get_cli_root_directory()]
        while len(directories) > 0:
            directory = directories.pop(0)

            try:
                project_config = self._project_config_manager.get_project_config(directory)
            except:
                continue
            if project_config and project_config.get("cloud-id", None) == cloud_id:
                return directory
            else:
                directories.extend(d for d in directory.iterdir() if d.is_dir())

        return False

    def get_source_files(self, directory: Path) -> List[Path]:
        """Returns the paths of all the source files in a directory.

        :param directory: the path to the directory to get the source files of
        :return: the list of source files in the given project directory
        """
        source_files = []

        for obj in directory.iterdir():
            if obj.is_dir():
                if (obj.name in ["bin", "obj", ".ipynb_checkpoints", "backtests", "live", "optimizations"] or
                        obj.name.startswith(".") or
                        # ignore python virtual environments
                        (obj / "pyvenv.cfg").is_file()):
                    continue

                source_files.extend(self.get_source_files(obj))

            if obj.suffix not in [".py", ".cs", ".ipynb", ".css", ".html"]:
                continue

            source_files.append(obj)

        return source_files

    def update_last_modified_time(self, local_file_path: Path, cloud_timestamp: datetime) -> None:
        """Updates the last modified time of a local path to that of the cloud counterpart.

        :param local_file_path: the path to the local file to update the last modified time of
        :param cloud_timestamp: the last modified time of the counterpart of the local file in the cloud
        """
        from os import utime
        from datetime import timezone

        # Timestamps are stored in UTC in the cloud, but utime() requires them in the local timezone
        time = cloud_timestamp.replace(tzinfo=timezone.utc).astimezone(tz=None)
        time = round(time.timestamp() * 1e9)
        utime(local_file_path, ns=(time, time))

    def copy_code(self, project_dir: Path, output_dir: Path) -> None:
        """Copies the source code of a project to another directory.

        :param project_dir: the directory of the project
        :param output_dir: the directory to copy the code to
        """
        from shutil import copyfile

        output_dir.mkdir(parents=True, exist_ok=True)

        for source_file in self.get_source_files(project_dir):
            target_file = output_dir / source_file.relative_to(project_dir)
            target_file.parent.mkdir(parents=True, exist_ok=True)
            copyfile(source_file, target_file)

    def create_new_project(self, project_dir: Path, language: QCLanguage) -> None:
        """Creates a new project directory and fills it with some useful files.

        :param project_dir: the directory of the new project
        :param language: the language of the new project
        """
        project_dir.mkdir(parents=True, exist_ok=True)

        project_config = self._project_config_manager.get_project_config(project_dir)
        project_config.set("algorithm-language", language.name)
        project_config.set("parameters", {})
        project_config.set("description", "")

        if language == QCLanguage.Python:
            self._generate_python_library_projects_config()
            self._generate_vscode_python_config(project_dir)
            self._generate_pycharm_config(project_dir)
        else:
            self._generate_vscode_csharp_config(project_dir)

            image = self._cli_config_manager.get_engine_image()
            framework_ver = self._docker_manager.get_image_label(image, 'target_framework',
                                                                 DEFAULT_LEAN_DOTNET_FRAMEWORK)
            self._generate_csproj(project_dir, framework_ver)
            self.generate_rider_config(project_dir)

    def delete_project(self, project_dir: Path) -> None:
        """Deletes a project directory.

        Raises an error if the project directory does not exist.

        :param project_dir: the directory of the project to delete
        """
        from shutil import rmtree
        try:
            rmtree(project_dir)
        except FileNotFoundError:
            raise RuntimeError(f"Failed to delete project. Could not find the specified path {project_dir}.")


    def get_local_project_path(self, project_name: str, cloud_id: Optional[int] = None, local_id: Optional[int] = None,
                               allow_corrupted: Optional[bool] = False) -> Path:
        """Returns the local path where a certain cloud project should be stored.

        If two cloud projects are named "Project", they are pulled to ./Project and ./Project 2.

        If you push a project with unsupported cloud name, a supported project name would be assigned.

        :param project_name: the cloud project name to get the project path of
        :param cloud_id: the cloud project id to get the project path of
        :param local_id: the cloud project local id to get the project path of
        :param allow_corrupted: true if a corrupted path can be used
        :return: the path to the local project directory
        """

        if cloud_id is not None and local_id is not None:
            raise ValueError("Cannot specify both cloud_id and local_id")

        if cloud_id is None and local_id is None:
            raise ValueError("Must specify either cloud_id or local_id")

        local_path = self._format_local_path(project_name)

        current_index = 1
        while True:
            # we first check the current project name
            path_suffix = "" if current_index == 1 else f" {current_index}"
            current_path = Path.cwd() / (local_path + path_suffix)

            if not current_path.exists():
                return current_path

            if cloud_id is not None:
                current_project_config = self._project_config_manager.get_project_config(current_path)
                if current_project_config.is_empty():
                    self._logger.error(f"'{current_path}' '{PROJECT_CONFIG_FILE_NAME}' file is corrupted!")
                    if allow_corrupted:
                        return current_path
                elif current_project_config.get("cloud-id") == cloud_id:
                    return current_path

            if local_id is not None:
                current_project_config = self._project_config_manager.get_project_config(current_path)
                if current_project_config.is_empty():
                    self._logger.error(f"'{current_path}' '{PROJECT_CONFIG_FILE_NAME}' file is corrupted!")
                    if allow_corrupted:
                        return current_path
                elif current_project_config.get("local-id") == local_id:
                    return current_path

            if current_index == 1:
                from re import findall

                ints_in_name = findall(r"(\s\d+)$", local_path)
                if len(ints_in_name) != 0:
                    # the current project name already exists, and it has a 'space + int' at the end of it
                    # so, we take that int and increment it
                    int_value = ints_in_name[0]
                    # we remove the int from the name because we will re add it in the loop
                    local_path = local_path[:-len(int_value)]
                    current_index = int(int_value)
            current_index += 1

    def rename_project_and_contents(self, old_path: Path, new_path: Path,) -> None:
        """Renames a project and updates the project config.

        :param old_path: the local project to rename
        :param new_path: the new path of the project
        """
        if not old_path.exists():
            raise RuntimeError(f"Failed to rename project. Could not find the specified path {old_path}.")
        if old_path == new_path:
            return
        from shutil import move
        move(old_path, new_path)
        self._rename_csproj_file(new_path)

    def get_projects_by_name_or_id(self, cloud_projects: List[QCProject],
                                   project: Optional[Union[str, int]]) -> List[QCProject]:
        """Returns a list of all the projects in the cloud that match the given name or id.

        :param cloud_projects: all projects fetched from the cloud
        :param project: the name or id of the project
        :return: a list of all the projects in the cloud that match the given name or id
        """
        search_by_id = isinstance(project, int)

        if project is not None:
            project_path = Path(project).as_posix() if not search_by_id else None
            projects = [cloud_project for cloud_project in cloud_projects
                        if (search_by_id and cloud_project.projectId == project or
                            not search_by_id and Path(cloud_project.name).as_posix() == project_path)]

            if len(projects) == 0:
                raise RuntimeError("No project with the given name or id exists in the cloud")
        else:
            projects = cloud_projects

        return projects

    def get_project_libraries(self, project_dir: Path) -> List[Path]:
        """Returns a list of all the libraries referenced by the given project.

        It will also recursively get all the libraries referenced by each library.

        :param project_dir: Path to the project
        :return List of all the libraries referenced by the given project
        """
        return self._get_project_libraries(project_dir)

    def _get_project_libraries(self, project_dir: Path, seen_projects: List[Path] = None) -> List[Path]:
        """Returns a list of all the libraries referenced by the given project.

        This is a helper method to recurse the libraries and get their dependencies as well.

        :param project_dir: Path to the project
        :param seen_projects: List of paths already seen, which serves as recursion stop criteria
        :return List of all the libraries referenced by the given project
        """
        if seen_projects is None:
            seen_projects = [project_dir]

        project_config = self._project_config_manager.get_project_config(project_dir)
        libraries_in_config = project_config.get("libraries", [])
        libraries = [LeanLibraryReference(**library).path.expanduser().resolve() for library in libraries_in_config]
        referenced_libraries = []

        for library_path in libraries:
            # Avoid infinite recursion
            if library_path in seen_projects:
                continue

            seen_projects.append(library_path)
            referenced_libraries.extend(self._get_project_libraries(library_path, seen_projects))
            referenced_libraries.append(library_path)

        return referenced_libraries

    def restore_csharp_project(self, csproj_file: Path, no_local: bool) -> None:
        """Restores a C# project if requested with the no_local flag and if dotnet is on the user's PATH.

        :param csproj_file: Path to the project's csproj file
        :param no_local: Whether restoring the packages locally must be skipped
        """
        from shutil import which
        from subprocess import run, STDOUT, PIPE
        from lean.models.errors import MoreInfoError

        if no_local:
            return

        if which("dotnet") is None:
            self._logger.info(f"Project {csproj_file.parent} will not be restored because dotnet was not found in PATH")
            return

        project_dir = csproj_file.parent
        self._logger.info(f"Restoring packages in '{self._path_manager.get_relative_path(project_dir)}' "
                          f"to provide local autocomplete")

        process = run(["dotnet", "restore", str(csproj_file)], cwd=project_dir, stdout=PIPE, stderr=STDOUT, text=True)
        self._logger.debug(process.stdout)

        if process.returncode != 0:
            raise RuntimeWarning("Something went wrong while restoring packages. "
                                 "You might be missing the .NET Core SDK in your dotnet installation. "
                                 "Local autocomplete functionality might be limited.")

        self._logger.info("Restored successfully")

    def try_restore_csharp_project(self, csproj_file: Path,
                                   original_csproj_content: Optional[str] = None,
                                   no_local: bool = False) -> None:
        """Restores a C# project if requested with the no_local flag and if dotnet is on the user's PATH.

        :param csproj_file: Path to the project's csproj file
        :param original_csproj_content: The original csproj file content
        :param no_local: Whether restoring the packages locally must be skipped
        """
        try:
            self.restore_csharp_project(csproj_file, no_local)
        except RuntimeWarning as e:
            if original_csproj_content is not None:
                self._logger.info(f"Reverting the changes to '{self._path_manager.get_relative_path(csproj_file)}'")
                csproj_file.write_text(original_csproj_content, encoding="utf-8")
            raise e


    def _format_local_path(self, cloud_path: str) -> str:
        """Converts the given cloud path into a local path which is valid for the current operating system.

        :param cloud_path: the path of the project in the cloud
        :return: the converted cloud_path so that it is valid locally
        """
        # Remove forbidden characters and OS-specific path separator that are not path separators on QuantConnect
        # Windows, \":*?"<>| are forbidden
        # Windows, \ is a path separator, but \ is not a path separator on QuantConnect
        # We follow the rules of windows for every OS

        for character in cloud_path:
            if self._path_manager.is_name_valid(character):
                continue
            cloud_path = cloud_path.replace(character, " ")

        # On Windows we need to ensure each path component is valid
        # We follow the rules of windows for every OS
        new_components = []

        for component in cloud_path.split("/"):
            # Some names are reserved
            for reserved_name in reserved_names:
                # If the component is a reserved name, we add an underscore to it so it can be used
                if component.upper() == reserved_name:
                    component += "_"

            # Components cannot start or end with a space
            component = component.strip(" ")

            # Components cannot end with a period
            component = component.rstrip(".")

            new_components.append(component)

        cloud_path = "/".join(new_components)

        return cloud_path


    def _generate_python_library_projects_config(self) -> None:
        """Generates the required configuration to enable autocomplete on Python library projects."""
        try:
            cli_root_dir = self._lean_config_manager.get_cli_root_directory()
        except:
            return
        from site import ENABLE_USER_SITE, getusersitepackages, getsitepackages

        library_dir = cli_root_dir / "Library"
        if not library_dir.is_dir():
            return

        if ENABLE_USER_SITE:
            site_packages_dir = getusersitepackages()
        else:
            site_packages_dir = getsitepackages()[0]

        self._generate_file(Path(site_packages_dir) / "lean-cli.pth", str(library_dir))

    def _generate_vscode_python_config(self, project_dir: Path) -> None:
        """Generates Python interpreter configuration and Python debugging configuration for VS Code.

        :param project_dir: the directory of the new project
        """
        from sys import executable
        from json import dumps

        self._generate_file(project_dir / ".vscode" / "settings.json", dumps({
            "python.pythonPath": executable,
            "python.languageServer": "Pylance"
        }, indent=4))

        launch_config = {
            "version": "0.2.0",
            "configurations": [
                {
                    "name": "Debug with Lean CLI",
                    "type": "python",
                    "request": "attach",
                    "connect": {
                        "host": "localhost",
                        "port": 5678
                    },
                    "pathMappings": [
                        {
                            "localRoot": "${workspaceFolder}",
                            "remoteRoot": "/LeanCLI"
                        }
                    ]
                }
            ]
        }

        try:
            library_dir = self._lean_config_manager.get_cli_root_directory() / "Library"
            launch_config["configurations"][0]["pathMappings"].append({
                "localRoot": str(library_dir),
                "remoteRoot": "/Library"
            })
        except:
            pass

        self._generate_file(project_dir / ".vscode" / "launch.json", dumps(launch_config, indent=4))

    def _generate_pycharm_config(self, project_dir: Path) -> None:
        """Generates Python interpreter configuration and Python debugging configuration for PyCharm.

        :param project_dir: the directory of the new project
        """
        from os import path

        # Generate Python JDK entry for PyCharm Professional and PyCharm Community
        for editor in ["PyCharm", "PyCharmCE"]:
            for directory in self._get_jetbrains_config_dirs(editor):
                self._generate_pycharm_jdk_entry(directory)

        self._generate_file(project_dir / ".idea" / f"{project_dir.name}.iml", """
<?xml version="1.0" encoding="UTF-8"?>
<module type="PYTHON_MODULE" version="4">
  <component name="NewModuleRootManager">
    <content url="file://$MODULE_DIR$" />
    <orderEntry type="jdk" jdkName="Lean CLI" jdkType="Python SDK" />
    <orderEntry type="sourceFolder" forTests="false" />
  </component>
</module>
        """)

        self._generate_file(project_dir / ".idea" / "misc.xml", """
<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="ProjectRootManager" version="2" project-jdk-name="Lean CLI" project-jdk-type="Python SDK" />
</project>
        """)

        self._generate_file(project_dir / ".idea" / "modules.xml", f"""
<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="ProjectModuleManager">
    <modules>
      <module fileurl="file://$PROJECT_DIR$/.idea/{project_dir.name}.iml" filepath="$PROJECT_DIR$/.idea/{project_dir.name}.iml" />
    </modules>
  </component>
</project>
        """)

        try:
            library_dir = self._lean_config_manager.get_cli_root_directory() / "Library"
            library_dir = f"$PROJECT_DIR$/{path.relpath(library_dir, project_dir)}".replace("\\", "/")
            library_mapping = f'<mapping local-root="{library_dir}" remote-root="/Library" />'
        except:
            library_mapping = ""

        self._generate_file(project_dir / ".idea" / "workspace.xml", f"""
<?xml version="1.0" encoding="UTF-8"?>
<project version="4">
  <component name="RunManager" selected="Python Debug Server.Debug with Lean CLI">
    <configuration name="Debug with Lean CLI" type="PyRemoteDebugConfigurationType" factoryName="Python Remote Debug">
      <module name="LEAN" />
      <option name="PORT" value="6000" />
      <option name="HOST" value="localhost" />
      <PathMappingSettings>
        <option name="pathMappings">
          <list>
            <mapping local-root="$PROJECT_DIR$" remote-root="/LeanCLI" />
            {library_mapping}
          </list>
        </option>
      </PathMappingSettings>
      <option name="REDIRECT_OUTPUT" value="true" />
      <option name="SUSPEND_AFTER_CONNECT" value="true" />
      <method v="2" />
    </configuration>
    <list>
      <item itemvalue="Python Debug Server.Debug with Lean CLI" />
    </list>
  </component>
</project>
        """)

    def _generate_pycharm_jdk_entry(self, pycharm_config_dir: Path) -> None:
        """Generates a "Lean CLI" Python JDK entry to PyCharm's internal JDK table.

        When we generate PyCharm's .idea directory we want to tell PyCharm where the Python interpreter is located.
        PyCharm stores this bit of configuration globally, so we find the global location and update it to our needs.

        If PyCharm is not installed yet, we create the configuration anyways.
        Once the user installs PyCharm, it will then automatically pick up the configuration we created in the past.

        :param pycharm_config_dir: the path to the global configuration directory of a PyCharm edition
        """
        from sys import path, executable, version_info
        # Parse the file containing PyCharm's internal table of Python interpreters
        jdk_table_file = pycharm_config_dir / "options" / "jdk.table.xml"
        if jdk_table_file.exists():
            root = self._xml_manager.parse(jdk_table_file.read_text(encoding="utf-8"))
        else:
            root = self._xml_manager.parse("""
<application>
  <component name="ProjectJdkTable">
  </component>
</application>
            """)

        # Don't do anything if the Lean CLI interpreter entry already exists
        if root.find(".//jdk/name[@value='Lean CLI']") is not None:
            return

        # Add the new JDK entry to the XML tree
        classpath_entries = [Path(p).as_posix() for p in path if p != "" and not p.endswith(".zip")]
        classpath_entries = [f'<root url="{p}" type="simple" />' for p in classpath_entries]
        classpath_entries = "\n".join(classpath_entries)

        component_element = root.find(".//component[@name='ProjectJdkTable']")
        component_element.append(self._xml_manager.parse(f"""
<jdk version="2">
  <name value="Lean CLI" />
  <type value="Python SDK" />
  <version value="Python {version_info.major}.{version_info.minor}.{version_info.micro}" />
  <homePath value="{Path(executable).as_posix()}" />
  <roots>
    <classPath>
      <root type="composite">
        {classpath_entries}
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/python-skeletons" type="simple" />
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/typeshed/stdlib/3" type="simple" />
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/typeshed/stdlib/2and3" type="simple" />
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/typeshed/third_party/3" type="simple" />
        <root url="file://$APPLICATION_HOME_DIR$/plugins/python/helpers/typeshed/third_party/2and3" type="simple" />
      </root>
    </classPath>
    <sourcePath>
      <root type="composite" />
    </sourcePath>
  </roots>
</jdk>
        """))

        # Save the modified XML tree
        self._generate_file(jdk_table_file, self._xml_manager.to_string(root))

    def _get_libraries_path_from_project(self, project_dir: Path) -> str:
        """Gets the libraries path from a given project.

        The returned path is relative to the project directory.
        - For a project in ${cli_root_dir}/Project, the returned path will be ../Library.
        - For a project in ${cli_root_dir}/SomeSubdirectory/Project, the returned path will be ../../Library.

        :param project_dir: The path to the project directory
        :return The path to the libraries directory relative to the project directory
        """
        cli_root_dir = self._lean_config_manager.get_cli_root_directory()
        project_dir_relative_to_cli = self._path_manager.get_relative_path(project_dir, cli_root_dir)

        return "../" * len(project_dir_relative_to_cli.parts) + "Library"

    def _generate_vscode_csharp_config(self, project_dir: Path) -> None:
        """Generates C# debugging configuration for VS Code.

        :param project_dir: the directory of the new project
        """
        self._generate_file(project_dir / ".vscode" / "launch.json", f"""
{{
    "version": "0.2.0",
    "configurations": [
        {{
            "name": "Debug with Lean CLI",
            "request": "attach",
            "type": "coreclr",
            "processId": "1",
            "pipeTransport": {{
                "pipeCwd": "${{workspaceRoot}}",
                "pipeProgram": "docker",
                "pipeArgs": ["exec", "-i", "lean_cli_vsdbg"],
                "debuggerPath": "/root/vsdbg/vsdbg",
                "quoteArgs": false
            }},
            "logging": {{
                "moduleLoad": false
            }},
            "sourceFileMap": {{
                "/Library": "${{workspaceFolder}}/{self._get_libraries_path_from_project(project_dir)}"
            }}
        }}
    ]
}}
        """)

    def _generate_csproj(self, project_dir: Path, framework_version: str) -> None:
        """Generates a .csproj file for the given project and returns the path to it.

        :param project_dir: the path of the new project
        """
        self._generate_file(project_dir / f"{project_dir.name}.csproj",
                            self.get_csproj_file_default_content(framework_version))

    def generate_rider_config(self, project_dir: Path) -> bool:
        """Generates C# debugging configuration for Rider.

        :param project_dir: the directory of the project
        :return: True if the configuration was generated successfully or changes where made, False if otherwise.
        """
        from pkg_resources import resource_string

        ssh_dir = Path("~/.lean/ssh").expanduser()

        # Add SSH keys to .lean/ssh if necessary
        ssh_dir.mkdir(parents=True, exist_ok=True)
        for name in ["key", "key.pub", "README.md"]:
            file_name = ssh_dir / name
            if not file_name.exists() or file_name.stat().st_size == 0:
                with (ssh_dir / name).open("wb+") as file:
                    file.write(resource_string("lean", f"ssh/{name}"))

        made_changes = False
        # Find Rider's global configuration directory for versions < 2022
        for directory in self._get_jetbrains_config_dirs("Rider"):
            directory_name_parts = directory.name.replace("Rider", "").split(".")  # e.g. "Rider2021.1"
            major_version = directory_name_parts[0]
            if not major_version or (major_version.isdigit() and int(major_version) < 2022):
                made_changes = self._generate_rider_below_v2022_debugger_entry(directory, ssh_dir)


        # For newer Rider versions (>= 2022), we can just create the ssh debugger configurations on a per-project basis
        rider_dir = project_dir / ".idea" / f".idea.{project_dir.stem}.dir" / ".idea"
        rider_dir.mkdir(parents=True, exist_ok=True)

        return self._generate_rider_above_v2022_debugger_entry(rider_dir, ssh_dir) or made_changes

    def _generate_rider_below_v2022_debugger_entry(self, rider_config_dir: Path, ssh_dir: Path) -> bool:
        """Generates a "root@localhost:2222" remote debugger entry to Rider's internal debugger configuration.

        If Rider is not installed yet, we create the configuration anyways.
        Once the user installs Rider, it will then automatically pick up the configuration we created in the past.

        :param rider_config_dir: the path to the global configuration directory of a Rider edition
        :param ssh_dir: the path to the directory containing the SSH keys
        :return: True if the configuration was generated successfully or changes where made, False if otherwise.
        """
        made_changes = False

        # Parse the file containing Rider's internal list of remote hosts
        debugger_file = rider_config_dir / "options" / "debugger.xml"
        if debugger_file.exists():
            root = self._xml_manager.parse(debugger_file.read_text(encoding="utf-8"))
        else:
            root = self._xml_manager.parse("""
    <application>
        <component name="XDebuggerSettings">
        </component>
    </application>
                """)
            made_changes = True

        component_element = root.find(".//component[@name='XDebuggerSettings']")

        if root.find(".//debuggers") is None:
            component_element.append(self._xml_manager.parse("<debuggers></debuggers>"))
            made_changes = True
        debuggers = root.find(".//debuggers")

        if debuggers.find(".//debugger[@id='dotnet_debugger']") is None:
            debuggers.append(self._xml_manager.parse('<debugger id="dotnet_debugger"></debugger>'))
            made_changes = True
        dotnet_debugger = debuggers.find(".//debugger[@id='dotnet_debugger']")

        if dotnet_debugger.find(".//configuration") is None:
            dotnet_debugger.append(self._xml_manager.parse("<configuration></configuration>"))
            made_changes = True
        configuration = dotnet_debugger.find(".//configuration")

        if configuration.find(".//option[@name='sshCredentials']") is None:
            configuration.append(self._xml_manager.parse('<option name="sshCredentials"></option>'))
            made_changes = True
        ssh_credentials = configuration.find(".//option[@name='sshCredentials']")

        required_value = f"&lt;credentials HOST=&quot;localhost&quot; PORT=&quot;2222&quot; USERNAME=&quot;root&quot; PRIVATE_KEY_FILE=&quot;{ssh_dir.as_posix()}/key&quot; USE_KEY_PAIR=&quot;true&quot; USE_AUTH_AGENT=&quot;false&quot; /&gt;"

        # Don't do anything if the required entry already exists
        for option in ssh_credentials.findall(f".//option"):
            if option.get("value") == required_value.replace("&lt;", "<").replace("&gt;", ">").replace("&quot;", '"'):
                return False

        # Add the new entry to the XML tree
        ssh_credentials.append(self._xml_manager.parse(f'<option value="{required_value}"/>'))

        # Save the modified XML tree
        self._generate_file(debugger_file, self._xml_manager.to_string(root))

        return made_changes

    def _generate_rider_above_v2022_debugger_entry(self, rider_config_dir: Path, ssh_dir: Path) -> bool:
        """Generates a "root@localhost:2222" ssh configuration for the debugger to attach to the remote process.

        :param rider_config_dir: the path to the project's rider configuration directory
        :param ssh_dir: the path to the directory containing the SSH keys
        :return: True if the configuration was generated successfully or changes where made, False if otherwise.
        """
        made_changes = False

        # Parse the file containing Rider's internal list of remote hosts
        ssh_configs_file = rider_config_dir / "sshConfigs.xml"
        if ssh_configs_file.exists():
            root = self._xml_manager.parse(ssh_configs_file.read_text(encoding="utf-8"))
        else:
            root = self._xml_manager.parse("""
<project version="4">
  <component name="SshConfigs">
    <configs>
    </configs>
  </component>
</project>
""")
            made_changes = True

        component_element = root.find(".//component[@name='SshConfigs']")

        if component_element.find(".//configs") is None:
            component_element.append(self._xml_manager.parse("<configs></configs>"))
            made_changes = True
        configs = root.find(".//configs")

        if configs.find(".//sshConfig[@id='dotnet_debugger']") is None:
            key_path = ssh_dir / "key"
            configs.append(self._xml_manager.parse(f'<sshConfig id="dotnet_debugger" host="localhost" port="2222" username="root" keyPath="{key_path}" useOpenSSHConfig="true" />'))
            made_changes = True

        # Save the modified XML tree
        self._generate_file(ssh_configs_file, self._xml_manager.to_string(root))

        return made_changes

    def _rename_csproj_file(self, project_path: Path) -> None:
        """Renames the csproj file in the project to name the project name.

        :param project_path: the local project path
        """
        csproj_file = next(project_path.glob("*.csproj"), None)
        if not csproj_file:
            return
        new_csproj_file = project_path / f'{project_path.name}.csproj'
        if new_csproj_file.exists():
            return
        from shutil import move
        move(csproj_file, new_csproj_file)

    def _generate_file(self, file: Path, content: str) -> None:
        """Writes to a file, which is created if it doesn't exist yet, and normalized the content before doing so.

        :param file: the file to write to
        :param content: the content to write to the file
        """
        file.parent.mkdir(parents=True, exist_ok=True)
        with file.open("w+", encoding="utf-8") as file:
            file.write(content.strip() + "\n")

    def _get_jetbrains_config_dirs(self, editor_name: str) -> List[Path]:
        """Returns the paths to the global configuration directories for all installed editions of a JetBrains IDE.

        :param editor_name: the name of the JetBrains IDE
        :return: the paths to the directories holding the global configuration for the specified IDE
        """
        # Find JetBrains' root directory containing the global configuration directories for all installed IDEs
        # See https://www.jetbrains.com/help/pycharm/project-and-ide-settings.html#ide_settings
        if self._platform_manager.is_host_windows():
            root_dir = Path("~/AppData/Roaming/JetBrains").expanduser()
        elif self._platform_manager.is_host_macos():
            root_dir = Path("~/Library/Application Support/JetBrains").expanduser()
        else:
            root_dir = Path("~/.config/JetBrains").expanduser()

        if not root_dir.exists():
            root_dir.mkdir(parents=True)

        # Find the global config directories for the given IDE
        directories = []

        for path in root_dir.iterdir():
            if not path.is_dir() or not path.name.startswith(editor_name):
                continue

            suffix = path.name.replace(editor_name, "")
            if len(suffix) > 0 and not suffix[0].isdigit():
                continue

            directories.append(path)

        if len(directories) == 0:
            directories.append(root_dir / editor_name)

        return directories

    def get_cloud_project_libraries(self,
                                    cloud_projects: List[QCProject],
                                    project: QCProject,
                                    seen_libraries: List[int] = None) -> Tuple[List[QCProject], List[QCProjectLibrary]]:
        """Gets the libraries referenced by the project and its dependencies from the given cloud projects.

        It recursively gets every Lean CLI library referenced by the project
        and the ones referenced by those libraries as well.

        :param cloud_projects: the cloud projects list to search in.
        :param project: the starting point for the libraries gathering.
        :param seen_libraries: list of seen library IDs to avoid infinite recursion.

        :return: two lists including the libraries referenced by the project.
            The first one containing the library projects that could be fetched and
            the second list containing the libraries that could not be fetched because the user has no access to them.
        """
        if seen_libraries is None:
            seen_libraries = []

        libraries = [cloud_project
                     for library in project.libraries
                     for cloud_project in cloud_projects if cloud_project.projectId == library.projectId]

        libraries_ids = [library.projectId for library in libraries]
        libraries_not_found = [library for library in project.libraries if library.projectId not in libraries_ids]

        referenced_libraries = []
        for library in libraries:
            # Avoid infinite recursion
            if library.projectId in seen_libraries:
                continue

            seen_libraries.append(library.projectId)
            libs, libs_not_found = self.get_cloud_project_libraries(cloud_projects, library, seen_libraries)
            referenced_libraries.extend(libs)
            libraries_not_found.extend(libs_not_found)

        libraries.extend(referenced_libraries)

        return list(set(libraries)), list(set(libraries_not_found))

    def get_cloud_projects_libraries(self,
                                     cloud_projects: List[QCProject],
                                     projects: List[QCProject],
                                     seen_projects: List[int] = None) -> Tuple[List[QCProject], List[QCProjectLibrary]]:
        """Gets the libraries referenced by the passed projects and its dependencies from the given cloud projects.

        It recursively gets every Lean CLI library referenced by the passed projects
        and the ones referenced by those libraries as well.

        :param cloud_projects: the cloud projects list to search in.
        :param projects: the starting point list of projects for the libraries gathering.
        :param seen_projects: list of seen project IDs to avoid infinite recursion.

        :return: two lists including the libraries referenced by the projects.
            The first one containing the library projects that could be fetched and
            the second list containing the libraries that could not be fetched because the user has no access to them.
        """
        if seen_projects is None:
            seen_projects = [project.projectId for project in projects]

        libraries = []
        libraries_not_found = []
        for project in projects:
            libs, libs_not_found = self.get_cloud_project_libraries(cloud_projects, project, seen_projects)
            libraries.extend(libs)
            libraries_not_found.extend(libs_not_found)

        return list(set(libraries)), list(set(libraries_not_found))

    @staticmethod
    def get_csproj_file_default_content(framework_version: str) -> str:
        return """
<Project Sdk="Microsoft.NET.Sdk">
    <PropertyGroup>
        <Configuration Condition=" '$(Configuration)' == '' ">Debug</Configuration>
        <Platform Condition=" '$(Platform)' == '' ">AnyCPU</Platform>
        <TargetFramework>{TARGET_FRAMEWORK}</TargetFramework>
        <OutputPath>bin/$(Configuration)</OutputPath>
        <AppendTargetFrameworkToOutputPath>false</AppendTargetFrameworkToOutputPath>
        <DefaultItemExcludes>$(DefaultItemExcludes);backtests/*/code/**;live/*/code/**;optimizations/*/code/**</DefaultItemExcludes>
        <NoWarn>CS0618</NoWarn>
    </PropertyGroup>
    <ItemGroup>
        <PackageReference Include="QuantConnect.Lean" Version="2.5.*"/>
        <PackageReference Include="QuantConnect.DataSource.Libraries" Version="2.5.*"/>
    </ItemGroup>
</Project>
        """.replace('{TARGET_FRAMEWORK}', framework_version)

    @staticmethod
    def get_csproj_file_path(project_dir: Path) -> Path:
        """Gets the path to the csproj file in the project directory.

        :param project_dir: Path to the project directory
        :return: Path to the csproj file in the project directory
        """
        return next((p for p in project_dir.iterdir() if p.name.endswith(".csproj")), None)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List
from lean.components.util.logger import Logger
from lean.models.json_module import JsonModule
from lean.models.logger import Option


def build_and_configure_modules(target_modules: List[str], module_list: List[JsonModule], organization_id: str,
                                lean_config: Dict[str, Any], properties: Dict[str, Any], logger: Logger,
                                environment_name: str):
    """Builds and configures the given modules

    :param target_modules: the requested modules
    :param module_list: the available modules
    :param organization_id: the organization id
    :param lean_config: the current lean configs
    :param properties: the user provided arguments
    :param logger: the logger instance
    :param environment_name: the environment name to use
    """
    for target_module_name in target_modules:
        module = non_interactive_config_build_for_name(lean_config, target_module_name, module_list, properties,
                                                       logger, environment_name)
        module.ensure_module_installed(organization_id)
        lean_config["environments"][environment_name].update(module.get_settings())


def non_interactive_config_build_for_name(lean_config: Dict[str, Any], target_module_name: str,
                                          module_list: List[JsonModule], properties: Dict[str, Any], logger: Logger,
                                          environment_name: str = None) -> JsonModule:
    return config_build_for_name(lean_config, target_module_name, module_list, properties, logger, interactive=False,
                                 environment_name=environment_name)


def find_module(target_module_name: str, module_list: List[JsonModule], logger: Logger) -> JsonModule:
    target_module: JsonModule = None
    # because we compare str we normalize everything to lower case
    target_module_name = target_module_name.lower()
    module_class_name = target_module_name.rfind('.')
    for module in module_list:
        # we search in the modules name and id
        module_id = module.get_id().lower()
        module_name = module.get_name().lower()

        if module_id == target_module_name or module_name == target_module_name:
            target_module = module
            break
        else:
            if (module_class_name != -1 and module_id == target_module_name[module_class_name + 1:]
                    or module_name == target_module_name[module_class_name + 1:]):
                target_module = module
                break

    if not target_module:
        for module in module_list:
            # we search in the modules configuration values, this is for when the user provides an environment
            if (module.is_value_in_config(target_module_name)
               or module_class_name != -1 and module.is_value_in_config(target_module_name[module_class_name + 1:])):
                target_module = module
        if not target_module:
            raise RuntimeError(f"""Failed to resolve module for name: '{target_module_name}'""")
    logger.debug(f'Found module \'{target_module_name}\' from given name')
    return target_module


def config_build_for_name(lean_config: Dict[str, Any], target_module_name: str, module_list: List[JsonModule],
                          properties: Dict[str, Any], logger: Logger, interactive: bool,
                          environment_name: str = None) -> JsonModule:
    target_module = find_module(target_module_name, module_list, logger)
    target_module.config_build(lean_config, logger, interactive=interactive, properties=properties,
                               environment_name=environment_name)
    _update_settings(logger, environment_name, target_module, lean_config)
    return target_module


def interactive_config_build(lean_config: Dict[str, Any], models: [JsonModule], logger: Logger,
                             user_provided_options: Dict[str, Any], show_secrets: bool, select_message: str,
                             multiple: bool, environment_name: str = None) -> [JsonModule]:
    """Interactively configures the brokerage to use.

    :param lean_config: the LEAN configuration that should be used
    :param models: the modules to choose from
    :param logger: the logger to use
    :param user_provided_options: the dictionary containing user provided options
    :param show_secrets: whether to show secrets on input
    :param select_message: the user facing selection message
    :param multiple: true if multiple selections are allowed
    :param environment_name: the target environment name
    :return: the brokerage the user configured
    """
    options = [Option(id=b, label=b.get_name()) for b in models]

    modules: [JsonModule] = []
    if multiple:
        modules = logger.prompt_list(select_message, options, multiple=True)
    else:
        module = logger.prompt_list(select_message, options, multiple=False)
        modules.append(module)

    for module in modules:
        module.config_build(lean_config, logger, interactive=True, properties=user_provided_options,
                            hide_input=not show_secrets, environment_name=environment_name)
        _update_settings(logger, environment_name, module, lean_config)
    if multiple:
        return modules
    return modules[-1]


def _update_settings(logger: Logger, environment_name: str, module: JsonModule,
                     lean_config: Dict[str, Any]) -> None:
    settings = module.get_settings()
    logger.debug(f'_update_settings({module}): Settings: {settings}')

    if environment_name:
        if "environments" not in lean_config:
            lean_config["environments"] = {}
        if environment_name not in lean_config["environments"]:
            lean_config["environments"][environment_name] = {}
        target = lean_config["environments"][environment_name]
    else:
        target = lean_config

    for key, value in settings.items():
        if key in target:
            from json import loads
            if isinstance(target[key], str) and target[key].startswith("["):
                # it already exists, and it's an array we need to merge
                logger.debug(f'_update_settings({module}): target[key]: {target[key]}')
                existing_value = loads(target[key])
                if value.startswith("["):
                    # the new value is also an array, merge them
                    existing_value = existing_value + loads(value)
                else:
                    existing_value.append(value)
                # guarantee order but uniqueness
                target[key] = list(dict.fromkeys(existing_value).keys())
            elif isinstance(target[key], list):
                # guarantee order but uniqueness
                target[key] = list(dict.fromkeys(target[key] + loads(value)).keys())
            else:
                target[key] = value
        else:
            target[key] = value




# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4
from os import path, getcwd, unlink, mkdir

from click import command, option, argument
from lean.click import LeanCommand
from lean.container import container


@command(cls=LeanCommand)
@argument("key", type=str, nargs=-1)
@option("--destination-folder", type=str, default="",
          help=f"The destination folder to download the object store values,"
               f" if not provided will use to current directory")
def get(key: [str], destination_folder: str):
    """
    Download an object store value to disk from the organization's cloud object store.

    :param key: The desired key to fetch, multiple can be provided.
    """
    organization_id = container.organization_manager.try_get_working_organization_id()
    api_client = container.api_client
    logger = container.logger

    logger.info(f"Fetching object store download url")
    url = api_client.object_store.get(key, organization_id, logger)
    if not destination_folder:
        destination_folder = getcwd()

    if not path.exists(destination_folder):
        mkdir(destination_folder)

    temp_file = path.join(destination_folder, f"{str(uuid4())}.zip")

    with logger.transient_progress() as progress:
        progress.add_task(f"Start downloading keys into {temp_file}:", total=None)
        logger.debug(f"Downloading: {url}")

        api_client.data.download_url(url, temp_file, lambda advance: None)

    logger.info(f"Unzipping object store keys values into: '{destination_folder}'")
    from zipfile import ZipFile
    with ZipFile(temp_file, 'r') as zip_ref:
        zip_ref.extractall(destination_folder)

    if path.exists(temp_file):
        logger.debug(f"Deleting temp file: '{temp_file}'")
        unlink(temp_file)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional

from click import command, option

from lean.click import LeanCommand, PathParameter
from lean.constants import PROJECT_CONFIG_FILE_NAME
from lean.container import container
from lean.models.encryption import ActionType

@command(cls=LeanCommand)
@option("--project",
        type=PathParameter(exists=True, file_okay=False, dir_okay=True),
        help="Path to the local project to push (all local projects if not specified)")
@option("--encrypt",
        is_flag=True, default=False,
        help="Push your local files and encrypt them before saving on the cloud")
@option("--decrypt",
        is_flag=True, default=False,
        help="Push your local files and decrypt them before saving on the cloud")
@option("--key",
              type=PathParameter(exists=True, file_okay=True, dir_okay=False),
              help="Path to the encryption key to use")
def push(project: Optional[Path], encrypt: Optional[bool], decrypt: Optional[bool], key: Optional[Path]) -> None:
    """Push local projects to QuantConnect.

    This command overrides the content of cloud files with the content of their respective local counterparts.

    This command will delete cloud files which don't have a local counterpart.
    """
    push_manager = container.push_manager
    encryption_action = None

    from lean.components.util.encryption_helper import validate_user_inputs_for_cloud_push_pull_commands
    validate_user_inputs_for_cloud_push_pull_commands(encrypt, decrypt, key)

    if encrypt:
        encryption_action = ActionType.ENCRYPT
    if decrypt:
        encryption_action = ActionType.DECRYPT

    # Parse which projects need to be pushed
    if project is not None:
        project_config_manager = container.project_config_manager
        project_config = project_config_manager.get_project_config(project)
        if not project_config.file.exists():
            raise RuntimeError(f"'{project}' is not a Lean project")

        if encrypt and key is not None:
            from lean.components.util.encryption_helper import validate_encryption_key_registered_with_cloud
            validate_encryption_key_registered_with_cloud(key, container.organization_manager, container.api_client) 

        push_manager.push_project(project, encryption_action, key)
    else:
        if key is not None:
            raise RuntimeError(f"Encryption key can only be specified when pushing a single project.")
        projects_to_push = [p.parent for p in Path.cwd().rglob(PROJECT_CONFIG_FILE_NAME)]
        push_manager.push_projects(projects_to_push, [], encryption_action, key)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from click import command, argument
from lean.click import LeanCommand
from lean.container import container


@command(cls=LeanCommand)
@argument("key", type=str)
def properties(key: str):
    """
    Get a value properties from the organization's cloud object store.

    :param key: The desired key to fetch the properties for.
    """
    organization_id = container.organization_manager.try_get_working_organization_id()
    api_client = container.api_client
    logger = container.logger
    data = api_client.object_store.properties(key, organization_id)

    try:
        headers = ["size", "modified", "key", "preview"]
        display_headers = ["Bytes", "Modified", "Filename", "Preview"]
        data_row = []
        for header in headers:
            if header == "preview":
                value = str(data["metadata"].get(header, "N/A"))
                data_row.append(_clean_up_preview(value))
            else:
                value = str(data["metadata"].get(header, ""))
                data_row.append(value)
        all_rows = [display_headers] + [data_row]
        column_widths = [max(len(row[i]) for row in all_rows) for i in range(len(all_rows[0]))]
        for row in all_rows:
            logger.info("  ".join(value.ljust(width) for value, width in zip(row, column_widths)))
    except KeyError as e:
        logger.error(f"Key {key} not found.")
    except Exception as e:
        logger.error(f"Error: {e}")


def _clean_up_preview(preview: str) -> str:
    return preview.rstrip()[:10]



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lean.components.api.api_client import *
from lean.components.util.logger import Logger


class ObjectStoreClient:
    """The ObjectStoreClient class contains methods to interact with object/* API endpoints."""

    def __init__(self, api_client: 'APIClient') -> None:
        """Creates a new ObjectStoreClient instance.

        :param api_client: the APIClient instance to use when making requests
        """
        self._api = api_client

    def properties(self, key: str, organization_id: str) -> str:
        """Returns the details of key from the object store.
        :param key: key of the object to retrieve
        :param organization_id: the id of the organization who's object store to retrieve from
        :return: the details of the specified object
        """
        payload = {
            "organizationId": organization_id,
            "key": key,
        }

        return self._api.post("object/properties", payload)

    def get(self, keys: [str], organization_id: str, logger: Logger) -> str:
        """Will fetch an url to download the requested keys
        :param keys: keys of the object to retrieve
        :param organization_id: the id of the organization who's object store to retrieve from
        :param logger: the logger instance to use
        :return: the url to download the requested keys
        """

        payload = {
            "organizationId": organization_id,
            "keys": keys,
        }

        data = self._api.post("object/get", payload)
        if "url" in data and data["url"]:
            # we got the url right away
            return data["url"]

        job_id = data["jobId"]
        if job_id:
            from time import sleep
            payload = {
                "organizationId": organization_id,
                "jobId": job_id,
            }

            with logger.transient_progress() as progress:
                progress.add_task("Waiting for files to be ready for download:", total=None)

                # we will retry up to 5 min to get the url
                retry_count = 0
                while not data["url"]:
                    sleep(3)
                    data = self._api.post("object/get", payload)
                    retry_count = retry_count + 1
                    if retry_count > ((60 * 5) / 3):
                        raise TimeoutError(f"Timeout waiting for object store job id {job_id}, please contact support")
        else:
            raise ValueError("JobId was not found in API response")

        return data["url"]

    def set(self, key: str, object_data: bytes, organization_id: str) -> None:
        """Sets the given key in the Object Store.
        :param key: key of the object to set
        :param object_data: the data to set
        :param organization_id: the id of the organization who's object store to set data in
        """
        payload = {
            "organizationId": organization_id,
            "key": key
        }
        extra_options = {
            "files": {"objectData": object_data}
        }
        self._api.post("object/set", payload, False, extra_options)

    def list(self, path: str, organization_id: str) -> str:
        """List all values for the given root key in the Object Store.

        :param path: root key for which to list all objects
        :param organization_id: the id of the organization who's object store to retrieve from
        :return: all objects for the given root key
        """
        payload = {
            "organizationId": organization_id,
            "path": path
        }

        data = self._api.post("object/list", payload)

        return data

    def delete(self, key: str, organization_id: str) -> None:
        """Deletes the given key from the Object Store.

        :param key: key of the object to delete
        :param organization_id: the id of the organization who's object store to delete from
        """
        payload = {
            "organizationId": organization_id,
            "key": key
        }

        self._api.post("object/delete", payload)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from click import command, argument

from lean.click import LeanCommand
from lean.container import container
from lean.models.errors import MoreInfoError


@command(cls=LeanCommand)
@argument("key", type=str)
def get(key: str) -> None:
    """Get the current value of a configurable option.

    Sensitive options like credentials cannot be retrieved this way for security reasons.
    Please open ~/.lean/credentials if you want to see your currently stored credentials.

    Run `lean config list` to show all available options.
    """
    cli_config_manager = container.cli_config_manager

    option = cli_config_manager.get_option_by_key(key)
    if option.is_sensitive:
        raise RuntimeError(
            "Sensitive options like credentials cannot be retrieved using `lean config get` for security reasons")

    value = option.get_value()
    if value is None:
        raise MoreInfoError(f"The option with key '{key}' doesn't have a value set",
                            "https://www.lean.io/docs/v2/lean-cli/api-reference/lean-config-set")

    logger = container.logger
    logger.info(value)



# QUANTCONNECT.COM - Democratizing Finance, Empowering Individuals.
# Lean CLI v1.0. Copyright 2021 QuantConnect Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from click import group, option, Context, pass_context, echo

from lean import __version__
from lean.click import verbose_option
from lean.components.util.click_aliased_command_group import AliasedCommandGroup
from lean.container import container
from lean.models.errors import MoreInfoError


@group(cls=AliasedCommandGroup, invoke_without_command=True)
@option("--version", is_flag=True, is_eager=True, help="Show the version and exit.")
@verbose_option()
@pass_context
def lean(ctx: Context, version: bool) -> None:
    """The Lean CLI by QuantConnect."""
    # This method is used as the command group for all `lean <command>` commands

    if ctx.invoked_subcommand is None:
        if version:
            program_name = ctx.find_root().info_name
            container.logger.info(f"{program_name} {__version__}")
            ctx.exit()
        else:
            echo(ctx.get_help())

    # Here we check whether `lean init` has already been called and this is an old CLI folder
    # before passing through the actual invoked command, to make sure this is not an old CLI folder.
    # TODO: This check is to be removed once some time has passed after
    #  the "one cli folder per organization" change is settled

    lean_config_manager = container.lean_config_manager
    try:
        lean_config_manager.get_lean_config()
    except MoreInfoError:
        # There is no config, proceed in case `lean init` is being invoked
        return

    # `lean init` was already called. Let's check whether this is an old Lean CLI directory
    organization_manager = container.organization_manager
    if organization_manager.get_working_organization_id() is None:
        raise RuntimeError(
            "This is an old Lean CLI root folder.\n"
            "From now on, a Lean CLI root folder must be created for each organization for improved usability.\n"
            "For each organization you'd like to use with the CLI, please create a new folder and run `lean init`.")
